"""
pyart.retrieve.echo_class
=========================

Functions for echo classification.

.. autosummary::
    :toctree: generated/

    steiner_conv_strat
    hydroclass_semisupervised
    data_for_centroids
    compute_centroids
    centroids_iter
    store_centroids
    select_samples
    make_platykurtic
    search_medoids
    split_clusters
    compute_intermediate_medoids
    determine_medoids
    synthetic_obs_distribution
    compute_ks_threshold
    compare_samples
    bell_function
    gaussian_function
    trapezoidal_function
    sample_bell
    sample_trapezoidal
    get_freq_band
    _standardize
    _destandardize
    _assign_to_class
    _assign_to_class_scan
    _compute_coeff_transform
    _get_mass_centers
    _mass_centers_table
    _data_limits_table
    _data_limits_centroids_table
    _bell_function_table
    _trapezoidal_function_table

"""

import traceback
from warnings import warn
from copy import deepcopy

import numpy as np
from scipy import interpolate
from scipy.stats import ks_2samp

from ..config import get_fillvalue, get_field_name, get_metadata
from ._echo_class import steiner_class_buff
from ..util import ma_broadcast_to

try:
    from sklearn_extra.cluster import KMedoids
    from sklearn.model_selection import train_test_split
    _SKLEARN_AVAILABLE = True
except ImportError:
    _SKLEARN_AVAILABLE = False

try:
    import dask
    _DASK_AVAILABLE = True
except ImportError:
    _DASK_AVAILABLE = False


def steiner_conv_strat(grid, dx=None, dy=None, intense=42.0,
                       work_level=3000.0, peak_relation='default',
                       area_relation='medium', bkg_rad=11000.0,
                       use_intense=True, fill_value=None,
                       refl_field=None):
    """
    Partition reflectivity into convective-stratiform using the Steiner et
    al. (1995) algorithm.

    Parameters
    ----------
    grid : Grid
        Grid containing reflectivity field to partition.
    dx, dy : float, optional
        The x- and y-dimension resolutions in meters, respectively. If None
        the resolution is determined from the first two axes values.
    intense : float, optional
        The intensity value in dBZ. Grid points with a reflectivity
        value greater or equal to the intensity are automatically
        flagged as convective. See reference for more information.
    work_level : float, optional
        The working level (separation altitude) in meters. This is the height
        at which the partitioning will be done, and should minimize bright band
        contamination. See reference for more information.
    peak_relation : 'default' or 'sgp', optional
        The peakedness relation. See reference for more information.
    area_relation : 'small', 'medium', 'large', or 'sgp', optional
        The convective area relation. See reference for more information.
    bkg_rad : float, optional
        The background radius in meters. See reference for more information.
    use_intense : bool, optional
        True to use the intensity criteria.
    fill_value : float, optional
         Missing value used to signify bad data points. A value of None
         will use the default fill value as defined in the Py-ART
         configuration file.
    refl_field : str, optional
         Field in grid to use as the reflectivity during partitioning. None
         will use the default reflectivity field name from the Py-ART
         configuration file.

    Returns
    -------
    eclass : dict
        Steiner convective-stratiform classification dictionary.

    References
    ----------
    Steiner, M. R., R. A. Houze Jr., and S. E. Yuter, 1995: Climatological
    Characterization of Three-Dimensional Storm Structure from Operational
    Radar and Rain Gauge Data. J. Appl. Meteor., 34, 1978-2007.

    """
    # Get fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # Parse field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')

    # parse dx and dy
    if dx is None:
        dx = grid.x['data'][1] - grid.x['data'][0]
    if dy is None:
        dy = grid.y['data'][1] - grid.y['data'][0]

    # Get coordinates
    x = grid.x['data']
    y = grid.y['data']
    z = grid.z['data']

    # Get reflectivity data
    ze = np.ma.copy(grid.fields[refl_field]['data'])
    ze = ze.filled(np.NaN)

    eclass = steiner_class_buff(ze, x, y, z, dx=dx, dy=dy, bkg_rad=bkg_rad,
                                work_level=work_level, intense=intense,
                                peak_relation=peak_relation,
                                area_relation=area_relation,
                                use_intense=use_intense,)

    return {'data': eclass.astype(np.int32),
            'standard_name': 'echo_classification',
            'long_name': 'Steiner echo classification',
            'valid_min': 0,
            'valid_max': 2,
            'comment_1': ('Convective-stratiform echo '
                          'classification based on '
                          'Steiner et al. (1995)'),
            'comment_2': ('0 = Undefined, 1 = Stratiform, '
                          '2 = Convective')}


def hydroclass_semisupervised(radar,
                              hydro_names=('AG', 'CR', 'LR', 'RP', 'RN', 'VI',
                                           'WS', 'MH', 'IH/HDG'),
                              var_names=('dBZ', 'ZDR', 'KDP', 'RhoHV',
                                         'H_ISO0'),
                              mass_centers=None,
                              weights=np.array([1., 1., 1., 0.75, 0.5]),
                              value=50., lapse_rate=-6.5, refl_field=None,
                              zdr_field=None, rhv_field=None, kdp_field=None,
                              temp_field=None, iso0_field=None,
                              hydro_field=None, entropy_field=None,
                              temp_ref='temperature', compute_entropy=False,
                              output_distances=False, vectorize=False):
    """
    Classifies precipitation echoes following the approach by Besic et al
    (2016).

    Parameters
    ----------
    radar : radar
        Radar object.
    hydro_names : array of str
        name of the types of hydrometeors
    var_names : array of str
        name of the variables
    mass_centers : ndarray 2D, optional
        The centroids for each variable and hydrometeor class in (nclasses,
        nvariables).
    weights : ndarray 1D, optional
        The weight given to each variable. Ordered by [dBZ, ZDR, KDP, RhoHV,
        H_ISO0]
    value : float
        The value controlling the rate of decay in the distance transformation
    lapse_rate : float
        The decrease in temperature for each vertical km [deg/km]
    refl_field, zdr_field, rhv_field, kdp_field, temp_field, iso0_field : str
        Inputs. Field names within the radar object which represent the
        horizonal reflectivity, the differential reflectivity, the copolar
        correlation coefficient, the specific differential phase, the
        temperature and the height respect to the iso0 fields. A value of None
        for any of these parameters will use the default field name as defined
        in the Py-ART configuration file.
    hydro_field : str
        Output. Field name which represents the hydrometeor class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature
        or height_over_iso0
    compute_entropy : bool
        If true, the entropy is computed
    output_distances : bool
        If true, the normalized distances to the centroids for each
        hydrometeor are provided as output
    vectorize : bool
        If true, a vectorized version of the class assignation is going to be
        used

    Returns
    -------
    fields_dict : dict
        Dictionary containing the retrieved fields

    References
    ----------
    Besic, N., Figueras i Ventura, J., Grazioli, J., Gabella, M., Germann, U.,
    and Berne, A.: Hydrometeor classification through statistical clustering
    of polarimetric radar measurements: a semi-supervised approach,
    Atmos. Meas. Tech., 9, 4425-4445, doi:10.5194/amt-9-4425-2016, 2016

    """
    # select the centroids as a function of frequency band
    if mass_centers is None:
        # assign coefficients according to radar frequency
        if 'frequency' in radar.instrument_parameters:
            mass_centers = _get_mass_centers(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            mass_centers = _mass_centers_table()['C']
            warn('Radar frequency unknown. ' +
                 'Default coefficients for C band will be applied')

    if hydro_field is None:
        hydro_field = get_field_name('radar_echo_classification')
    if compute_entropy:
        if entropy_field is None:
            entropy_field = get_field_name('hydroclass_entropy')

    # Get the data fields
    fields_dict = {}
    for var_name in var_names:
        if var_name == 'dBZ':
            if refl_field is None:
                refl_field = get_field_name('reflectivity')
            radar.check_field_exists(refl_field)
            fields_dict.update({
                var_name: deepcopy(radar.fields[refl_field]['data'])})
        elif var_name == 'ZDR':
            if zdr_field is None:
                zdr_field = get_field_name('differential_reflectivity')
            radar.check_field_exists(zdr_field)
            fields_dict.update({
                var_name: deepcopy(radar.fields[zdr_field]['data'])})
        elif var_name == 'KDP':
            if kdp_field is None:
                kdp_field = get_field_name('specific_differential_phase')
            radar.check_field_exists(kdp_field)
            fields_dict.update({
                var_name: deepcopy(radar.fields[kdp_field]['data'])})
        elif var_name == 'RhoHV':
            if rhv_field is None:
                rhv_field = get_field_name('cross_correlation_ratio')
            radar.check_field_exists(rhv_field)
            fields_dict.update({
                var_name: deepcopy(radar.fields[rhv_field]['data'])})
        elif var_name == 'H_ISO0':
            if temp_ref == 'temperature':
                if temp_field is None:
                    temp_field = get_field_name('temperature')
                radar.check_field_exists(temp_field)
                # convert temp in relative height respect to iso0
                temp = deepcopy(radar.fields[temp_field]['data'])
                fields_dict.update({var_name: temp*(1000./lapse_rate)})
            else:
                if iso0_field is None:
                    iso0_field = get_field_name('height_over_iso0')
                radar.check_field_exists(iso0_field)
                fields_dict.update({
                    var_name: deepcopy(radar.fields[iso0_field]['data'])})
        else:
            raise ValueError(
                'Variable ' + var_name + ' unknown. '
                + 'Valid variable names for hydrometeor classification are: '
                + 'H_ISO0, dBZ, ZDR, KDP and RhoHV')

    # standardize data and centroids
    mc_std = np.empty(
        np.shape(mass_centers), dtype=fields_dict[var_names[0]].dtype)
    for i, var_name in enumerate(var_names):
        mc_std[:, i] = _standardize(mass_centers[:, i], var_name)
        fields_dict[var_name] = _standardize(fields_dict[var_name], var_name)

    # if entropy has to be computed get transformation parameters
    t_vals = None
    if compute_entropy:
        t_vals = _compute_coeff_transform(
            mc_std, weights=weights, value=value)

    # assign to class
    if vectorize:
        hydroclass_data, entropy_data, prop_data = _assign_to_class_scan(
            fields_dict, mc_std, var_names=var_names, weights=weights,
            t_vals=t_vals)
    else:
        hydroclass_data, entropy_data, prop_data = _assign_to_class(
            fields_dict, mc_std, weights=weights, t_vals=t_vals)

    # prepare output fields
    fields_dict = dict()
    hydro = get_metadata(hydro_field)
    hydro['data'] = hydroclass_data
    hydro.update({'_FillValue': 0})
    labels = ['NC']
    ticks = [1]
    boundaries = [0.5, 1.5]
    for i, hydro_name in enumerate(hydro_names):
        labels.append(hydro_name)
        ticks.append(i+2)
        boundaries.append(i+2.5)
    hydro.update({
        'labels': labels,
        'ticks': ticks,
        'boundaries': boundaries})
    fields_dict.update({'hydro': hydro})

    if compute_entropy:
        entropy = get_metadata(entropy_field)
        entropy['data'] = entropy_data
        fields_dict.update({'entropy': entropy})

        if output_distances:
            for hydro_name in hydro_names:
                field_name = 'proportion_'+field_name
                prop = get_metadata(field_name)
                prop['data'] = prop_data[:, :, i]
                fields_dict.update({field_name: prop})

    return fields_dict


def data_for_centroids(radar, lapse_rate=-6.5, refl_field=None,
                       zdr_field=None, rhv_field=None, kdp_field=None,
                       temp_field=None, iso0_field=None,
                       temp_ref='temperature', nsamples_max=20000):
    """
    Prepares the data to compute the centroids of the hydrometeor
    classification

    Parameters
    ----------
    radar : radar
        Radar object.
    lapse_rate : float
        The decrease in temperature for each vertical km [deg/km]
    refl_field, zdr_field, rhv_field, kdp_field, temp_field, iso0_field : str
        Inputs. Field names within the radar object which represent the
        horizonal reflectivity, the differential reflectivity, the copolar
        correlation coefficient, the specific differential phase, the
        temperature and the height respect to the iso0 fields. A value of None
        for any of these parameters will use the default field name as defined
        in the Py-ART configuration file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature
        or height_over_iso0
    nsamples_max : int
        Maximum number of samples to keep from each radar volume

    Returns
    -------
    refl_std, zdr_std, kdp_std, rhohv_std, relh_std : 1D-array
        The standardized valid data

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if rhv_field is None:
        rhv_field = get_field_name('cross_correlation_ratio')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')

    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
    else:
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')

    # extract fields and parameters from radar
    radar.check_field_exists(refl_field)
    radar.check_field_exists(zdr_field)
    radar.check_field_exists(rhv_field)
    radar.check_field_exists(kdp_field)
    if temp_ref == 'temperature':
        radar.check_field_exists(temp_field)
    else:
        radar.check_field_exists(iso0_field)

    refl = deepcopy(radar.fields[refl_field]['data'])
    zdr = deepcopy(radar.fields[zdr_field]['data'])
    rhohv = deepcopy(radar.fields[rhv_field]['data'])
    kdp = deepcopy(radar.fields[kdp_field]['data'])

    if temp_ref == 'temperature':
        # convert temp in relative height respect to iso0
        temp = deepcopy(radar.fields[temp_field]['data'])
        relh = temp*(1000./lapse_rate)
    else:
        relh = deepcopy(radar.fields[iso0_field]['data'])

    # filter data out of limits
    dlimits_dict = _data_limits_centroids_table()

    mx, mn = dlimits_dict['dBZ']
    refl[(refl < mn) | (refl > mx)] = np.ma.masked

    mx, mn = dlimits_dict['ZDR']
    zdr[(zdr < mn) | (zdr > mx)] = np.ma.masked

    mx, mn = dlimits_dict['RhoHV']
    rhohv[(rhohv < mn) | (rhohv > mx)] = np.ma.masked

    mx, mn = dlimits_dict['KDP']
    kdp[(kdp < mn) | (kdp > mx)] = np.ma.masked

    mx, mn = dlimits_dict['H_ISO0']
    relh[(relh < mn) | (relh > mx)] = np.ma.masked

    # keep only gates with all values valid
    mask = np.ma.getmaskarray(refl)
    mask = np.logical_or(mask, np.ma.getmaskarray(zdr))
    mask = np.logical_or(mask, np.ma.getmaskarray(rhohv))
    mask = np.logical_or(mask, np.ma.getmaskarray(kdp))
    mask = np.logical_or(mask, np.ma.getmaskarray(relh))
    valid = np.logical_not(mask)

    refl = refl[valid]
    zdr = zdr[valid]
    rhohv = rhohv[valid]
    kdp = kdp[valid]
    relh = relh[valid]

    # keep only nsamples_max values of the volume
    if refl.size <= nsamples_max:
        return refl, zdr, kdp, rhohv, relh

    rg = np.random.default_rng(seed=0)
    ind = rg.integers(low=0, high=refl.size, size=nsamples_max)

    refl = refl[ind]
    zdr = zdr[ind]
    rhohv = rhohv[ind]
    kdp = kdp[ind]
    relh = relh[ind]

    return refl, zdr, kdp, rhohv, relh


def compute_centroids(features_matrix, weight=(1., 1., 1., 1., 0.75),
                      var_names=('dBZ', 'ZDR', 'KDP', 'RhoHV', 'H_ISO0'),
                      hydro_names=('AG', 'CR', 'LR', 'RP', 'RN', 'VI', 'WS',
                                   'MH', 'IH/HDG'),
                      nsamples_iter=20000, external_iterations=30,
                      internal_iterations=10, alpha=0.01, cv_approach=True,
                      num_samples_arr=(30, 35, 40), n_samples_syn=50,
                      nmedoids_min=1, acceptance_threshold=0.5, band='C',
                      relh_slope=0.001, parallelized=False, sample_data=True,
                      kmax_iter=100, keep_labeled_data=True):
    """
    Given a features matrix computes the centroids

    Parameters
    ----------
    features_matrix : 2D-array
        matrix of size (nsamples, nvariables)
    weight : tuple
        Weight given to each feature in the KS test
    var_names : tupple
        List of name variables
    hydro_names : tupple
        List of hydrometeor types
    nsamples_iter : int
        Number of samples of the features matrix in each external iteration
    external_iterations : int
        Number of iterations of the external loop. This number will determine
        how many medoids are computed for each hydrometeor class.
    internal_iterations : int
        Maximum number of iterations of the internal loop
    acceptance_threshold : float
        Threshold on the inter-quantile coefficient of dispersion of the
        medoids above which the medoid of the class is not acceptable.
    alpha : float
        Minimum value to accept the cluster according to p
    cv_approach : bool
        If true it is used a critical value approach to reject or accept
        similarity between observations and reference. If false it is used a
        p-value approach
    num_samples_arr : 1D-array
        Array containing the possible number of observation samples to use when
        comparing with reference
    n_samples_syn : int
        Number of samples from reference used in comparison
    nmedoids_min : int
        Minimum number of valid intermediate medoids to compute a final medoid
    band : str
        Frequency band of the radar data. Can be C, S or X
    relh_slope : float
        The slope used to transform the height relative to the iso0 into
        a sigmoid function.
    parallelized : bool
        If True the processing is going to be parallelized
    sample_data : bool
        If True the data is going to be sampled at each external loop
    kmax_iter : int
        Maximum number of iterations of the kmedoids algorithm
    keep_labeled_data : bool
        If True the labeled data is going to be kept.

    Returns
    -------
    labeled_data : 2D-array
        matrix of size (nsamples, nvariables) containing the observations
    labels : 1D-array
        array with the labels index
    medoids_dict : dict
        Dictionary containing the intermediate medoids for each hydrometeor
        type
    final_medoids_dict : dict
        Dictionary containing the final medoids for each hydrometeor type

    """
    if not _SKLEARN_AVAILABLE:
        warn(
            'Unable to compute centroids. scikit-learn package not available')
        return None, None, dict(), None

    if parallelized:
        if not _DASK_AVAILABLE:
            warn('dask not available: The processing will not be parallelized')
            parallelized = False

    labels = None
    labeled_data = None
    medoids_dict = dict()
    if not parallelized:
        rg = np.random.default_rng(seed=0)
        for i in range(external_iterations):
            new_labels, new_labeled_data, inter_medoids_dict = centroids_iter(
                features_matrix, i, rg, weight=weight, var_names=var_names,
                hydro_names=hydro_names, nsamples_iter=nsamples_iter,
                external_iterations=external_iterations,
                internal_iterations=internal_iterations, alpha=alpha,
                cv_approach=cv_approach, num_samples_arr=num_samples_arr,
                n_samples_syn=n_samples_syn, band=band, relh_slope=relh_slope,
                sample_data=sample_data, kmax_iter=kmax_iter,
                keep_labeled_data=keep_labeled_data)
            if not inter_medoids_dict:
                continue

            labels, labeled_data, medoids_dict = store_centroids(
                new_labels, new_labeled_data, inter_medoids_dict, hydro_names,
                labels=labels, labeled_data=labeled_data,
                medoids_dict=medoids_dict,
                keep_labeled_data=keep_labeled_data)
    else:
        jobs = []
        for i in range(external_iterations):
            rg = np.random.default_rng(seed=None)
            features_matrix_aux = dask.delayed(features_matrix)
            jobs.append(dask.delayed(centroids_iter)(
                features_matrix_aux, i, rg, weight=weight, var_names=var_names,
                hydro_names=hydro_names, nsamples_iter=nsamples_iter,
                external_iterations=external_iterations,
                internal_iterations=internal_iterations, alpha=alpha,
                cv_approach=cv_approach, num_samples_arr=num_samples_arr,
                n_samples_syn=n_samples_syn, band=band, relh_slope=relh_slope,
                sample_data=sample_data, kmax_iter=kmax_iter,
                keep_labeled_data=keep_labeled_data))
        try:
            jobs = dask.compute(*jobs)

            for (i, (new_labels, new_labeled_data,
                     inter_medoids_dict)) in enumerate(jobs):
                if keep_labeled_data:
                    if new_labels is None:
                        nlabeled = 0
                    else:
                        nlabeled = new_labels.size
                    print(
                        'iteration '+str(i+1)+' labeled data '+str(nlabeled))
                if not inter_medoids_dict:
                    continue
                labels, labeled_data, medoids_dict = store_centroids(
                    new_labels, new_labeled_data, inter_medoids_dict,
                    hydro_names, labels=labels, labeled_data=labeled_data,
                    medoids_dict=medoids_dict,
                    keep_labeled_data=keep_labeled_data)
            del jobs
        except Exception as ee:
            warn(str(ee))
            traceback.print_exc()

    if not medoids_dict:
        warn('Data could not be labeled')
        return None, None, dict(), None

    final_medoids_dict = determine_medoids(
        medoids_dict, var_names, hydro_names, nmedoids_min=nmedoids_min,
        acceptance_threshold=acceptance_threshold, kmax_iter=kmax_iter)

    return labeled_data, labels, medoids_dict, final_medoids_dict


def centroids_iter(features_matrix, iteration, rg,
                   weight=(1., 1., 1., 1., 0.75),
                   var_names=('dBZ', 'ZDR', 'KDP', 'RhoHV', 'H_ISO0'),
                   hydro_names=('AG', 'CR', 'LR', 'RP', 'RN', 'VI', 'WS',
                                'MH', 'IH/HDG'),
                   nsamples_iter=20000, external_iterations=30,
                   internal_iterations=10, alpha=0.01, cv_approach=True,
                   num_samples_arr=(30, 35, 40), n_samples_syn=50,
                   band='C', relh_slope=0.001, sample_data=True,
                   kmax_iter=100, keep_labeled_data=True):
    """
    External iteration of the centroids computation

    Parameters
    ----------
    features_matrix : 2D-array
        matrix of size (nsamples, nvariables)
    iteration : int
        iteration number
    rg : Random Generator
        Random generator
    weight : tuple
        Weight given to each feature in the KS test
    var_names : tupple
        List of name variables
    hydro_names : tupple
        List of hydrometeor types
    nsamples_iter : int
        Number of samples of the features matrix in each external iteration
    external_iterations : int
        Number of iterations of the external loop. This number will determine
        how many medoids are computed for each hydrometeor class.
    internal_iterations : int
        Maximum number of iterations of the internal loop
    alpha : float
        Minimum value to accept the cluster according to p
    cv_approach : bool
        If true it is used a critical value approach to reject or accept
        similarity between observations and reference. If false it is used a
        p-value approach
    num_samples_arr : 1D-array
        Array containing the possible number of observation samples to use when
        comparing with reference
    n_samples_syn : int
        Number of samples from reference used in comparison
    band : str
        Frequency band of the radar data. Can be C, S or X
    relh_slope : float
        The slope used to transform the height relative to the iso0 into
        a sigmoid function.
    sample_data : Bool
        If True the feature matrix will be sampled
    kmax_iter : int
        Maximum number of iterations of the k-medoids algorithm
    keep_labeled_data : Bool
        If True the labeled data is kept.

    Returns
    -------
    labeled_data : 2D-array
        matrix of size (nsamples, nvariables) containing the observations
    labels : 1D-array
        array with the labels index
    medoids_dict : dict
        Dictionary containing the intermediate medoids for each hydrometeor
        type
    final_medoids_dict : dict
        Dictionary containing the final medoids for each hydrometeor type

    """
    print('\n\n\nExternal loop. Iteration '+str(iteration+1)+'/' +
          str(external_iterations))
    # external loop to identify clusters
    ks_threshold, n_samples = compute_ks_threshold(
        rg, alpha=alpha, n_samples_syn=n_samples_syn,
        num_samples_arr=num_samples_arr)
    synthetic_obs = synthetic_obs_distribution(
        rg, var_names, hydro_names, band=band, relh_slope=relh_slope)

    if nsamples_iter > features_matrix.shape[0]:
        warn('Number of samples lower than number of samples per iteration')
        fm_sample = deepcopy(features_matrix)
    elif not sample_data:
        fm_sample = deepcopy(features_matrix)
    else:
        fm_sample, _ = train_test_split(
            features_matrix, train_size=nsamples_iter)

    # Uses sklearn.metrics.pairwise_distances for the metric
    # metric can be also those of scipy.spatial.distance
    kmedoids = KMedoids(
        n_clusters=len(hydro_names), metric='seuclidean', method='alternate',
        init='k-medoids++', max_iter=kmax_iter, random_state=None).fit(
            fm_sample)

    new_labels, new_labeled_data, _ = search_medoids(
        fm_sample, kmedoids.labels_, synthetic_obs, var_names, hydro_names,
        weight, ks_threshold, alpha, cv_approach, n_samples_syn, n_samples, 1,
        iteration_max=internal_iterations, relh_slope=relh_slope,
        kmax_iter=kmax_iter)

    if new_labels is None:
        print('No data labeled in internal loops')
        return None, None, dict()

    print('labeled data in internal loop: '+str(new_labels.size))

    # Compute medoids as the median of the clustered data
    inter_medoids_dict = compute_intermediate_medoids(
        new_labeled_data, new_labels, hydro_names, kmax_iter=kmax_iter)
    if not keep_labeled_data:
        new_labels = None
        new_labeled_data = None

    return new_labels, new_labeled_data, inter_medoids_dict


def store_centroids(new_labels, new_labeled_data, inter_medoids_dict,
                    hydro_names, labels=None, labeled_data=None,
                    medoids_dict=None, keep_labeled_data=True):
    """
    Store the centroids data to its respective recipients

    Parameters
    ----------
    new_labels : 1D-array
        array with the labels index
    new_labeled_data : 2D-array
        matrix of size (nsamples, nvariables) containing the observations
    inter_medoids_dict : dict
        Dictionary containing the intermediate medoids for each hydrometeor
        type
    hydro_names : array of str
        The name of the hydrometeor types
    labels : 1D-array or None
        array where to store the new labels
    labeled_data : 2D-array or None
        matrix of size (nsamples, nvariables) where to store the labeled data
    medoids_dict : dict
        Dictionary where to store the new medoids
    keep_labeled_data : Bool
        If True the labeled data is going to be stored

    Returns
    -------
    labels : 1D-array or None
        array with the labels index
    labeled_data : 2D-array or None
        matrix of size (nsamples, nvariables) containing the observations
    medoids_dict : dict
        Dictionary containing the intermediate medoids for each hydrometeor
        type

    """
    # store the correctly identified data and its labels
    if keep_labeled_data:
        if labels is None:
            labels = deepcopy(new_labels)
            labeled_data = deepcopy(new_labeled_data)
        else:
            labels = np.append(labels, new_labels, axis=-1)
            labeled_data = np.append(
                labeled_data, new_labeled_data, axis=0)
    del new_labels
    del new_labeled_data

    # store the new medoids
    for hydro_name in hydro_names:
        if hydro_name in inter_medoids_dict:
            if hydro_name in medoids_dict:
                medoids_dict[hydro_name] = np.append(
                    medoids_dict[hydro_name],
                    inter_medoids_dict[hydro_name], axis=0)
            else:
                medoids_dict.update({
                    hydro_name: inter_medoids_dict[hydro_name]})

    if labels is not None:
        print('total labeled data: '+str(labels.size))

    return labels, labeled_data, medoids_dict


def select_samples(fm, rg, nbins=110, pdf_zh_max=20000, pdf_relh_max=10000,
                   sigma_zh=0.75, sigma_relh=1.5, randomize=True,
                   platykurtic_dBZ=True, platykurtic_H_ISO0=True):
    """
    Selects the data to be used to compute the centroids

    Parameters
    ----------
    fm : 2D array (nsamples, nfeatures)
        Initial data
    rg : Random Generator
        The random generator used to randomize the polarimetric variables
    nbins : int
        Number of bins of the histograms used to make the data platykurtik
    pdf_zh_max : int
        Multiplication factor to the Guassian function over the
        reflectivity that determines the number of samples for each bin
    pdf_relh_max : int
        Multiplication factor to the Guassian function over the
        height relative to the iso-0 that determines the number of samples for
        each bin
    sigma_zh, sigma_relh : float
        sigma of the respective Gaussian functions
    randomize : bool
        if True the quantized data is randomized
    platykurtic_dBZ : Bool
        If True makes the reflectivity distribution platykurtic
    platykurtic_H_ISO0 : Bool
        If True makes the height relative to the iso-0 platykurtic

    Returns
    -------
    fm_sample : 2D array (nsamples, nfeatures)
        The selected data

    """
    nsamples = fm.shape[0]
    if randomize:
        print('Randomizing data')
        nfeatures = fm.shape[1]
        for i in range(nfeatures-1):
            vals = np.unique(fm[:, i])
            step = np.median(vals[1:]-vals[:-1])
            print('Number of unique values before randomization: {}'.format(
                vals.shape))
            print('vmin: {} vmax: {}'.format(vals.min(), vals.max()))
            print('Step between values: {}'.format(step))
            fm[:, i] += rg.random(nsamples)*step-step/2.
            print('Number of unique values after randomization: {}'.format(
                fm[:, i].shape))

    refl, zdr, kdp, rhohv, relh = make_platykurtic(
        fm[:, 0], fm[:, 1], fm[:, 2], fm[:, 3], fm[:, 4],
        nbins=nbins, pdf_zh_max=pdf_zh_max, pdf_relh_max=pdf_relh_max,
        sigma_zh=sigma_zh, sigma_relh=sigma_relh,
        platykurtic_dBZ=platykurtic_dBZ,
        platykurtic_H_ISO0=platykurtic_H_ISO0)

    fm_sample = np.transpose(np.array([refl, zdr, kdp, rhohv, relh]))

    print('Selected '+str(fm_sample.shape[0])+' samples out of ' +
          str(nsamples))

    return fm_sample


def make_platykurtic(refl, zdr, kdp, rhohv, relh, nbins=110,
                     pdf_zh_max=20000, pdf_relh_max=10000, sigma_zh=0.75,
                     sigma_relh=1.5, platykurtic_dBZ=True,
                     platykurtic_H_ISO0=True):
    """
    Prepares the data to compute the centroids of the hydrometeor
    classification

    Parameters
    ----------
    refl, zdr, kdp, rhohv, relh : 1D array
        Arrays containing the data
    nbins : int
        Number of bins of the histogram for each dataset
    pdf_zh_max : int
        Multiplication factor to the Guassian function over the
        reflectivity that determines the number of samples for each bin
    pdf_relh_max : int
        Multiplication factor to the Guassian function over the
        height relative to the iso-0 that determines the number of samples for
        each bin
    sigma_zh, sigma_relh : float
        sigma of the respective Gaussian functions
    platykurtic_dBZ : Bool
        If True makes the reflectivity distribution platykurtic
    platykurtic_H_ISO0 : Bool
        If True makes the height relative to the iso-0 platykurtic

    Returns
    -------
    refl, zdr, kdp, rhohv, relh : 1D array
        The selected data

    """
    x_vals = np.linspace(-1.1, 1.1, num=nbins)

    if platykurtic_dBZ:
        print('Making dBz distribution platykurtic')
        # make reflectivity platykurtik
        _, bin_edges = np.histogram(refl, bins=nbins)
        pdf = np.array(
            gaussian_function(x_vals, mu=0., sigma=sigma_zh, normal=True) *
            pdf_zh_max, dtype=int)

        refl_aux = []
        zdr_aux = []
        rhohv_aux = []
        kdp_aux = []
        relh_aux = []
        for i in range(nbins):
            ind = np.where(
                (refl >= bin_edges[i]) & (refl < bin_edges[i+1]))[0]
            if ind.size > pdf[i]:
                ind = ind[:pdf[i]]

            refl_aux.extend(refl[ind])
            zdr_aux.extend(zdr[ind])
            rhohv_aux.extend(rhohv[ind])
            kdp_aux.extend(kdp[ind])
            relh_aux.extend(relh[ind])

        refl = np.array(refl_aux)
        zdr = np.array(zdr_aux)
        rhohv = np.array(rhohv_aux)
        kdp = np.array(kdp_aux)
        relh = np.array(relh_aux)

    if platykurtic_H_ISO0:
        # Make relative height platykurtik
        print('Making H_ISO0 distribution platykurtic')
        _, bin_edges = np.histogram(relh, bins=nbins)

        pdf = np.array(
            gaussian_function(x_vals, mu=0., sigma=sigma_relh, normal=True) *
            pdf_relh_max, dtype=int)

        refl_aux = []
        zdr_aux = []
        rhohv_aux = []
        kdp_aux = []
        relh_aux = []
        for i in range(nbins):
            ind = np.where(
                (relh >= bin_edges[i]) & (relh < bin_edges[i+1]))[0]
            if ind.size > pdf[i]:
                ind = ind[:pdf[i]]

            refl_aux.extend(refl[ind])
            zdr_aux.extend(zdr[ind])
            rhohv_aux.extend(rhohv[ind])
            kdp_aux.extend(kdp[ind])
            relh_aux.extend(relh[ind])

        refl = np.array(refl_aux)
        zdr = np.array(zdr_aux)
        rhohv = np.array(rhohv_aux)
        kdp = np.array(kdp_aux)
        relh = np.array(relh_aux)

    return refl, zdr, kdp, rhohv, relh


def search_medoids(fm, clust_labels, synthetic_obs, var_names, hydro_names,
                   weight, ks_threshold, alpha, cv_approach, n_samples_syn,
                   n_samples, iteration, iteration_max=10, relh_slope=0.001,
                   kmax_iter=100):
    """
    Given a features matrix computes the centroids. This function is recursive

    Parameters
    ----------
    fm : 2D-array
        feature matrix of size (nsamples, nvariables)
    clust_labels : 1D-array
        cluster labels
    synthetic_obs : dict of dicts
        dictionary of type dict[var_name][hydro_name] with the samples
        corresponding to each membership function distribution
    var_names : list
        variables names
    hydro_names : list
        hydrometeor types names
    weight : Tupple
        weight assigned to each variable when looking at the similarity
        between membership functions pdfs and real observations
    ks_threshold : float
        Treshold used in the Kolmogorov–Smirnov comparison. If a statistic is
        above this value the similarity is rejected
    alpha : float
        Parameter alpha
    cv_approach : bool
        If true it is used a critical value approach to reject or accept
        similarity between observations and reference. If false it is used a
        p-value approach
    n_samples_syn : int
        Number of samples of the synthetic observations used in the KS test
    n_samples : int
        Number of real observations used to compare with the membership
        functions
    iteration : int
        index of the current iteration in the internal loop
    iteration_max : int
        Maximum number of iterations allowed in the internal loop
    relh_slope : float
        slope of the sigmoid function used to standardize height relative to
        the iso-0 data into
    kmax_iter : int
        Maximum number of iterations of the k-medoids algorithm

    Returns
    -------
    new_labels : 1D-array of int or None
        Array containing the label of each new clustered observation. None if
        no observation has been clustered
    new_labeled_data : 2D_array of floats or None.
        Array containing the clustered observations. None if no observation
        has been clustered
    iteration : int
        current iteration

    """
    print('\n\nInternal loop. Iteration: '+str(iteration)+'/' +
          str(iteration_max))

    if iteration >= iteration_max:
        warn('Maximum number of iterations reached or not enough samples')
        return None, None, iteration

    (hydro_labels, labeled_data, cluster_labels,
     nonlabeled_data) = compare_samples(
        var_names, hydro_names, weight, synthetic_obs, fm, clust_labels,
        ks_threshold, alpha, cv_approach, n_samples,
        n_samples_syn=n_samples_syn, relh_slope=relh_slope)

    n_labeled = 0
    if hydro_labels is not None:
        n_labeled = hydro_labels.size
    print('iteration: '+str(iteration)+' labeled points: '+str(n_labeled) +
          '/'+str(fm.shape[0]))

    n_nonlabeled = 0
    if cluster_labels is not None:
        n_nonlabeled = cluster_labels.size
    if n_nonlabeled <= n_samples:
        return hydro_labels, labeled_data, iteration_max

    # split each non-identified cluster into two and compare with reference
    clusters = np.unique(cluster_labels)
    print(str(clusters.size)+' clusters are not valid. ' +
          'Splitting invalid clusters')

    iteration += 1
    iteration1 = np.empty(clusters.size)
    iteration2 = np.empty(clusters.size)
    for icluster, cluster_id in enumerate(clusters):
        fm1, clust_labels1, fm2, clust_labels2 = split_cluster(
            nonlabeled_data, cluster_labels, cluster_id, n_samples,
            kmax_iter=kmax_iter)

        if fm1 is None:
            iteration1[icluster] = iteration_max
            hydro_labels1 = None
        else:
            (hydro_labels1, labeled_data1,
             iteration1[icluster]) = search_medoids(
                fm1, clust_labels1, synthetic_obs, var_names, hydro_names,
                weight, ks_threshold, alpha, cv_approach, n_samples_syn,
                n_samples, iteration, iteration_max=iteration_max,
                relh_slope=relh_slope, kmax_iter=kmax_iter)

        if fm2 is None:
            iteration2[icluster] = iteration_max
            hydro_labels2 = None
        else:
            (hydro_labels2, labeled_data2,
             iteration2[icluster]) = search_medoids(
                fm2, clust_labels2, synthetic_obs, var_names, hydro_names,
                weight, ks_threshold, alpha, cv_approach, n_samples_syn,
                n_samples, iteration, iteration_max=iteration_max,
                relh_slope=relh_slope, kmax_iter=kmax_iter)

        if hydro_labels1 is not None:
            # add the data
            if hydro_labels is None:
                hydro_labels = hydro_labels1
                labeled_data = labeled_data1
            else:
                hydro_labels = np.append(hydro_labels, hydro_labels1, axis=-1)
                labeled_data = np.append(labeled_data, labeled_data1, axis=0)

        if hydro_labels2 is not None:
            # add the data
            if hydro_labels is None:
                hydro_labels = hydro_labels2
                labeled_data = labeled_data2
            else:
                hydro_labels = np.append(hydro_labels, hydro_labels2, axis=-1)
                labeled_data = np.append(labeled_data, labeled_data2, axis=0)

    if (np.all(iteration1 >= iteration_max) and
            np.all(iteration2 >= iteration_max)):
        iteration = iteration_max

    return hydro_labels, labeled_data, iteration


def split_cluster(fm, labels, icluster, n_samples, kmax_iter=100):
    """
    Splits the elements of a features matrix corresponding to cluster icluster
    into 2 using the k-medoids algorithm

    Parameters
    ----------
    fm : 2D-array of floats
        feature matrix of size (nsamples, nvariables)
    labels : 1D-array
        labels corresponding to each sample
    icluster : int
        ID of the cluster to split
    n_samples : int
        minimum number of samples to consider the new set as valid
    kmax_iter : int
        Maximum number of iterations in the k-medoids algorithm

    Returns
    -------
    fm1, fm2 : 2D-array of floats or None
        The feature matrix corresponding tot he new sets. None if the new set
        is too small
    clust_labels1, clust_labels2 : 1D_array of ints or None.
        labels of the new split

    """
    ind_cluster = np.where(labels == icluster)[0]
    fm_cluster = fm[ind_cluster, :]

    kmedoids = KMedoids(
            n_clusters=2, metric='seuclidean', method='alternate',
            init='k-medoids++', max_iter=kmax_iter, random_state=None).fit(
                fm_cluster)
    ind1 = np.where(kmedoids.labels_ == 0)[0]
    ind2 = np.where(kmedoids.labels_ == 1)[0]

    print('\nCluster ID: '+str(icluster))
    print('Number of samples in fm1: '+str(ind1.size))
    print('Number of samples in fm2: '+str(ind2.size))

    # check if the number of samples is too small to proceed
    if ind1.size < n_samples:
        print('Number of non-labeled samples left ('+str(ind1.size) +
              ') smaller than number of samples necessary for clustering (' +
              str(int(n_samples))+')')
        fm1 = None
        clust_labels1 = None
    else:
        fm1 = fm_cluster[ind1, :]
        clust_labels1 = kmedoids.labels_[ind1]

    # check if the number of samples is too small to proceed
    if ind2.size < n_samples:
        print('Number of non-labeled samples left ('+str(ind2.size) +
              ') smaller than number of samples necessary for clustering (' +
              str(int(n_samples))+')')
        fm2 = None
        clust_labels2 = None
    else:
        fm2 = fm_cluster[ind2, :]
        clust_labels2 = kmedoids.labels_[ind2]

    return fm1, clust_labels1, fm2, clust_labels2


def compute_intermediate_medoids(fm, labels, hydro_names, kmax_iter=100):
    """
    Computes the intermediate medoids from the labeled data

    Parameters
    ----------
    fm : 2D-array of floats
        The features matrix
    labels : 1D-array of ints
        The label of each sample
    hydro_names : 1D-array of str
        Name of the hydrometeors
    kmax_iter : int
        Maximum number of iterations of the kmedoids algorithm

    Returns
    -------
    inter_medoids_dict : dict
        dictionary with the medoids for each hydrometeor class

    """
    inter_medoids_dict = dict()
    for ihydro, hydro_name in enumerate(hydro_names):
        ind = np.where(labels == ihydro)[0]
        if ind.size == 0:
            continue
        # medoids = np.median(fm[ind, :], axis=0, keepdims=True)
        kmedoids = KMedoids(
            n_clusters=1, metric='seuclidean', method='alternate',
            init='k-medoids++', max_iter=kmax_iter, random_state=None).fit(
                fm[ind, :])
        inter_medoids_dict.update({hydro_name: kmedoids.cluster_centers_})

    return inter_medoids_dict


def determine_medoids(medoids_dict, var_names, hydro_names, nmedoids_min=1,
                      acceptance_threshold=0.5, kmax_iter=100):
    """
    Computes the final medoids from the medoids found at each iteration

    Parameters
    ----------
    medoids_dict : dict
        dictionary of type dict[hydro_name]= array(niterations, nvars)
        containing the medoids for each hydrometeor type computed at each
        iteration
    var_names : 1D-array of str
        Name of variables
    hydro_names : 1D-array of str
        Name of hydrometeors
    nmedoids_min : int
        Minimum number of intermediate medoids to compute an intermediate
        medoid
    acceptance_threshold : float
        Threshold on the inter-quantile coefficient of dispersion of the
        medoids above which the medoid of the class is not acceptable.

    Returns
    -------
    final_medoids_dict : dict
        dictionary of type dict[hydro_name]= medoids_arr containing the
        medoids for each hydrometeor type that has been accepted

    """
    final_medoids_dict = dict()
    nvars = len(var_names)
    for hydro_name in hydro_names:
        if hydro_name not in medoids_dict:
            warn('No intermediate medoids found for class {}'.format(
                hydro_name))
            continue
        coef = 0.
        medoids = medoids_dict[hydro_name]
        if medoids.shape[0] < nmedoids_min:
            warn('Not enough intermediate medoids for class {}'.format(
                hydro_name))
            continue
        for ivar, var_name in enumerate(var_names):
            medoid_var = deepcopy(medoids[:, ivar])
            # shift the distribution towards positive values
            min_val = medoid_var.min()
            if min_val < 0:
                warn('Distribution of variable {} for hydrometeor type {} has negative values.'.format(
                    var_name, hydro_name))
                medoid_var -= min_val
            quant75 = np.quantile(medoid_var, 0.75)
            quant25 = np.quantile(medoid_var, 0.25)
            if quant75+quant25 == 0.:
                if quant75 == 0.:
                    pass
                else:
                    warn('Inter-quantile cannot be computed')
                    nvars -= 1
            else:
                cqv = (quant75-quant25)/(quant75+quant25)
                if cqv <= 0:
                    warn('Variable {} has negative inter-quantile {}'.format(
                        var_name, cqv))
                    nvars -= 1
                else:
                    coef += cqv
        coef = coef/nvars
        if coef > acceptance_threshold:
            warn('Inter-quantile coefficient of dispersion (' +
                 str(coef)+') larger than treshold (' +
                 str(acceptance_threshold) +
                 ') no valid centroids for class '+hydro_name)
            continue
        print(
            'medoids for '+hydro_name +
            ' found. Inter-quantile coefficient of dispersion: ' +
            str(coef))
        kmedoids = KMedoids(
            n_clusters=1, metric='seuclidean', method='alternate',
            init='k-medoids++', max_iter=kmax_iter, random_state=None).fit(
                medoids)
        final_medoids_dict.update({
            hydro_name: np.squeeze(kmedoids.cluster_centers_)})
        # final_medoids_dict.update({hydro_name: np.median(medoids, axis=0)})
    return final_medoids_dict


def synthetic_obs_distribution(rg, var_names, hydro_names, band='C',
                               relh_slope=0.001, par_var=0.05):
    """
    Gets the samples corresponding to the theoretical probability density
    function of each hydrometeor and variable

    Parameters
    ----------
    rg : Random generator
        The random generator used to generate samples of a uniform
        distribution
    var_names : 1D-array
        Name of the variables
    hydro_names : 1D-array
        Name of the hydrometeor types
    nsamples : int
        Number of samples for each distribution
    band : str
        Frequency band of the radar data. Can be C, S or X
    relh_slope : float
        The slope used to transform the height relative to the iso0 into
        a sigmoid function.
    par_var : float
        Variability of each parameter of the membership function.

    Returns
    -------
    synthetic_obs : dict of dicts
        dictionary of type dict[var_name][hydro_name] with the samples
        corresponding to each distribution

    """
    uniform_samples = rg.random(size=1000000, dtype=np.float32)

    synthetic_obs = dict()
    for var_name in var_names[:-1]:
        var_dict = dict()
        m_arr = np.empty(len(hydro_names))
        a_arr = np.empty(len(hydro_names))
        b_arr = np.empty(len(hydro_names))
        for ihydro, hydro_name in enumerate(hydro_names):
            # transform the uniform distribution according to the bell-shaped
            # distribution with random variations
            m, a, b = _bell_function_table()[band][var_name][hydro_name]
            min_m, max_m = [m-par_var*m, m+par_var*m]
            min_a, max_a = [a-par_var*a, a+par_var*a]
            min_b, max_b = [b-par_var*b, b+par_var*b]

            m_arr[ihydro] = (max_m-min_m)*rg.random()+min_m
            a_arr[ihydro] = (max_a-min_a)*rg.random()+min_a
            b_arr[ihydro] = (max_b-min_b)*rg.random()+min_b
        mn = np.min(m_arr-1.5*a_arr)
        mx = np.max(m_arr+1.5*a_arr)

        for ihydro, hydro_name in enumerate(hydro_names):
            pdf_samples = sample_bell(
                m=m_arr[ihydro], a=a_arr[ihydro], b=b_arr[ihydro], mn=mn,
                mx=mx)(uniform_samples)
            if var_name == 'RhoHV':
                pdf_samples[pdf_samples > 1.] = 1.
            var_dict.update({hydro_name: pdf_samples})
        synthetic_obs.update({var_name: var_dict})

    # synthetic observation for height relative to iso0
    var_dict = dict()
    v1_arr = np.empty(len(hydro_names))
    v2_arr = np.empty(len(hydro_names))
    v3_arr = np.empty(len(hydro_names))
    v4_arr = np.empty(len(hydro_names))
    for ihydro, hydro_name in enumerate(hydro_names):
        # transform the uniform distribution according to the trapezoidal
        # distribution with random variations
        v1, v2, v3, v4 = _trapezoidal_function_table()[hydro_name]
        min_v1, max_v1 = [v1-par_var*v1, v1+par_var*v1]
        min_v2, max_v2 = [v2-par_var*v2, v2+par_var*v2]
        min_v3, max_v3 = [v3-par_var*v3, v3+par_var*v3]
        min_v4, max_v4 = [v4-par_var*v4, v3+par_var*v4]

        v1_rand = (max_v1-min_v1)*rg.random()+min_v1
        v2_rand = (max_v2-min_v2)*rg.random()+min_v2
        v3_rand = (max_v3-min_v3)*rg.random()+min_v3
        v4_rand = (max_v4-min_v4)*rg.random()+min_v4

        if v1_rand < v2_rand < v3_rand < v4_rand:
            v1 = v1_rand
            v2 = v2_rand
            v3 = v3_rand
            v4 = v4_rand
        v1_arr[ihydro] = v1
        v2_arr[ihydro] = v2
        v3_arr[ihydro] = v3
        v4_arr[ihydro] = v4
    mn = np.min(v1_arr)
    mx = np.max(v4_arr)
    for ihydro, hydro_name in enumerate(hydro_names):
        pdf_samples = sample_trapezoidal(
            v1=v1_arr[ihydro], v2=v2_arr[ihydro], v3=v3_arr[ihydro],
            v4=v4_arr[ihydro], mn=mn, mx=mx)(uniform_samples)

        # standardize the samples of the distribution
        pdf_samples = _standardize(
            pdf_samples, var_names[-1], relh_slope=relh_slope)
        var_dict.update({hydro_name: pdf_samples})
    synthetic_obs.update({var_names[-1]: var_dict})

    return synthetic_obs


def compute_ks_threshold(rg, alpha=0.01, n_samples_syn=50,
                         num_samples_arr=(30, 35, 40)):
    """
    Computes the threshold of acceptance for the Kolmogorov–Smirnov test

    Parameters
    ----------
    rg : Random generator
        The random generator used to generate samples of a uniform
        distribution
    alpha : float
        test significance (alpha)
    n_samples_syn : int
        Number of samples from reference used to compare
    num_samples_arr : 1D-array
        Number of observation samples used to compare with reference to choose
        from

    Returns
    -------
    ks_threshold : float
        Critical value
    num_samples : float
        The selected number of samples

    """
    num_samples = rg.choice(num_samples_arr)
    ks_threshold = np.sqrt(
        (-np.log(alpha/2.)*(num_samples+n_samples_syn) /
         (2.*num_samples*n_samples_syn)))

    return ks_threshold, num_samples


def compare_samples(var_names, hydro_names, weight, synthetic_obs, fm,
                    clust_labels, ks_threshold, alpha, cv_approach,
                    n_samples_obs, n_samples_syn=50, margin_ratio=0.1,
                    relh_slope=0.001):
    """
    Compares the distribution of the clustered samples with the expected
    distribution

    Parameters
    ----------
    var_names : 1D-array of str
        Name of the variables
    hydro_names : 1D-array of str
        Name of the hydrometeor types
    weight : dict
        Weight given to each variable
    synthetic_obs : dict of dicts
        dictionary of type dict[var_name][hydro_name] with the samples
        corresponding to each distribution
    fm : 2D-array
        Features matrix (nsamples, nvars)
    clust_labels : 1D-array of nsamples
        Labels of each sample
    ks_threshold : float
        Threshold to accept the cluster as similar to the synthetic
        distribution
    alpha : float
        parameter alpha
    cv_approach : bool
        If true it is used a critical value approach to reject or accept
        similarity between observations and reference. If false it is used a
        p-value approach
    n_samples_obs : int
        Number of observations used in the KS test
    n_samples_syn : int
        Number of samples of the synthetic observations used in the KS test
    margin_ratio : float
        Percentage over range of observations margin to make the synthetic
        observations local
    relh_slope : float
        The slope used to transform the height relative to the iso0 into
        a sigmoid function.

    Returns
    -------
    labels : 1D-array
        array with the index of the label for the labeled data and -1 for that
        not labeled

    """
    total_weight = np.sum(weight)
    labels = -1+np.zeros(clust_labels.size, dtype=np.uint8)
    hydro_names_aux = list(hydro_names)
    for jhydro in np.unique(clust_labels):
        best_stat = 1e6
        best_p = -1.
        for ihydro, hydro_name in enumerate(hydro_names_aux):
            total_stat = 0.
            total_p = 0.
            for ivar, var_name in enumerate(var_names):
                # select the MF and the real observations
                so_aux = synthetic_obs[var_name][hydro_name]
                real_obs = deepcopy(fm[:, ivar])
                if var_name == 'H_ISO0':
                    real_obs = _standardize(
                        real_obs, 'H_ISO0', relh_slope=relh_slope)
                real_obs = real_obs[clust_labels == jhydro]

                # check if there are enough samples in cluster
                if real_obs.size <= n_samples_obs:
                    total_stat = total_weight
                    total_p = total_weight
                    continue

                # give the MF the same limits as the real obs
                obs_rng_min = real_obs.min()
                obs_rng_max = real_obs.max()
                margin = (obs_rng_max-obs_rng_min)*margin_ratio
                so_rng_min = obs_rng_min-margin
                so_rng_max = obs_rng_max+margin
                so_aux = so_aux[
                    (so_aux >= so_rng_min) & (so_aux <= so_rng_max)]

                # compare the real observations with the synthetic data
                if so_aux.size <= n_samples_syn:
                    statistic = 1.
                    p = 0.
                else:
                    # sampling period of observations
                    isamp_obs = int(np.ceil(real_obs.size/n_samples_obs))
                    real_obs = real_obs[::isamp_obs]
                    isamp_syn = int(np.ceil(so_aux.size/n_samples_syn))
                    so_aux = so_aux[::isamp_syn]
                    statistic, p = ks_2samp(
                        so_aux, real_obs, alternative='two-sided',
                        mode='auto')
                total_stat += statistic*weight[ivar]
                total_p += p*weight[ivar]
            total_stat /= total_weight
            total_p /= total_weight

            # check if data pass the test and is better than previous score
            if ((cv_approach and (total_stat < ks_threshold)) or
                    (not cv_approach and (total_p > alpha))):
                if ((cv_approach and (total_stat < best_stat)) or
                        (not cv_approach and (total_p > best_p))):
                    jhydro_aux = jhydro
                    ihydro_aux = ihydro
                    hydro_name_aux = hydro_name
                    best_stat = total_stat
                    best_p = total_p

        if best_stat < 1e6 or best_p > -1.:
            labels[clust_labels == jhydro_aux] = ihydro_aux
            print(
                'test passed for variable '+hydro_name_aux +
                ' with total statistic '+str(best_stat) +
                ' and required statistic '+str(ks_threshold)+' or total p ' +
                str(best_p)+' and required alpha '+str(alpha))
            hydro_names_aux.remove(hydro_name_aux)

    ind_id = np.where(labels > -1)[0]
    ind_noid = np.where(labels == -1)[0]

    hydro_labels = None
    labeled_data = None
    if ind_id.size > 0:
        hydro_labels = labels[ind_id]
        labeled_data = fm[ind_id, :]

    cluster_labels = None
    nonlabeled_data = None
    if ind_noid.size > 0:
        cluster_labels = clust_labels[ind_noid]
        nonlabeled_data = fm[ind_noid, :]

    return hydro_labels, labeled_data, cluster_labels, nonlabeled_data


def bell_function(x_vals, m=39., a=19., b=10.):
    """
    Bell-shaped probability function

    Parameters
    ----------
    x_vals : float or array
        the x values
    m, a, b : float
        Parameters describing the bell-shaped function. The default are those
        of reflectivity in rain at C-band. The input data is in dBZ

    Returns
    -------
    prob : float or array
        The probability values given the x_vals

    """
    y = 1./(1.+np.power(np.abs((x_vals-m)/a), 2.*b))
    y[~np.isfinite(y)] = 0.
    return y


def gaussian_function(x_vals, mu=25., sigma=19., normal=True):
    """
    Bell-shaped probability function

    Parameters
    ----------
    x_vals : float or array
        the x values
    m, a, b : float
        Parameters describing the bell-shaped function. The default are those
        of reflectivity in rain at C-band. The input data is in dBZ

    Returns
    -------
    prob : float or array
        The probability values given the x_vals

    """
    y = np.exp(-1*np.power(x_vals-mu, 2.)/(2*np.power(sigma, 2.)))
    if normal:
        y *= 1./(sigma*np.sqrt(2.*np.pi))
    return y


def trapezoidal_function(x_vals, v1=-2500., v2=-2200, v3=-300, v4=0):
    """
    trapezoidal probability function

    Parameters
    ----------
    x_vals : float or array
        the x values
    v1, v2, v3, v4 : float
        Parameters describing the trapezoidal function. The default are those
        of rain. The input data is in m respect to the iso0

    Returns
    -------
    prob : float or array
        The probability values given the x_vals

    """
    prob = np.zeros(x_vals.shape)
    prob[(x_vals > v1) & (x_vals <= v2)] = (
        (x_vals[(x_vals > v1) & (x_vals <= v2)]-v1)/(v2-v1))
    prob[(x_vals > v2) & (x_vals <= v3)] = 1.
    prob[(x_vals > v3) & (x_vals <= v4)] = (
        (v4-x_vals[(x_vals > v3) & (x_vals <= v4)])/(v4-v3))

    return prob


def sample_bell(m=39., a=19., b=10., mn=-10., mx=60.):
    """
    returns the function that computes the inverse version of the normalized
    cumulative sum of a Bell-shaped function

    Parameters
    ----------
    m, a, b : float
        Parameters describing the bell-shaped function. The default are those
        of reflectivity in rain at C-band. The input data is in dBZ
    mn, mx : float
        Range of the data

    Returns
    -------
    inverse_cdf : func
        Inverse cumulative distribution function

    """
    x = np.linspace(mn, mx, num=200)
    y = bell_function(x, m=m, a=a, b=b)  # probability density function, pdf
    cdf_y = np.abs(np.cumsum(y+1e-10))  # cumulative distribution func, cdf
    cdf_y = cdf_y/cdf_y.max()       # takes care of normalizing cdf to 1.0
    inverse_cdf = interpolate.interp1d(
        cdf_y, x, fill_value='extrapolate')  # this is a function
    return inverse_cdf


def sample_trapezoidal(v1=-2500., v2=-2200, v3=-300, v4=0, mn=-5000.,
                       mx=5000.):
    """
    returns the function that computes the inverse version of the normalized
    cumulative sum of a Trapezoidal distribution

    Parameters
    ----------
    v1, v2, v3, v4 : float
        Parameters describing the Trapezoidal function. The default are those
        of rain. The input data is in m respect to the iso-0
    mn, mx : float
        Range of the data

    Returns
    -------
    inverse_cdf : func
        Inverse cumulative distribution function

    """
    x = np.linspace(mn, mx, num=100)
    y = trapezoidal_function(
        x, v1=v1, v2=v2, v3=v3, v4=v4)  # pdf
    cdf_y = np.abs(np.cumsum(y+1e-10))  # cdf
    cdf_y = cdf_y/cdf_y.max()  # takes care of normalizing cdf to 1.0
    inverse_cdf = interpolate.interp1d(
        cdf_y, x, fill_value='extrapolate')  # this is a function
    return inverse_cdf


def get_freq_band(freq):
    """
    Returns the frequency band name (S, C, X, ...).

    Parameters
    ----------
    freq : float
        Radar frequency [Hz].

    Returns
    -------
    freq_band : str
        Frequency band name.

    """
    if 2e9 <= freq < 4e9:
        return 'S'
    if 4e9 <= freq < 8e9:
        return 'C'
    if 8e9 <= freq <= 12e9:
        return 'X'

    warn('Unknown frequency band')

    return None


def _standardize(data, field_name, mx=None, mn=None, relh_slope=0.005):
    """
    Streches the radar data to -1 to 1 interval.

    Parameters
    ----------
    data : array
        Radar field.
    field_name : str
        Type of field (H_ISO0, dBZ, ZDR, KDP or RhoHV).
    mx, mn : floats or None, optional
        Data limits for array values.
    relh_slope : float
        The slope used to transform the relative height into a sigmoid
        function

    Returns
    -------
    field_std : dict
        Standardized radar data.

    """
    if field_name == 'H_ISO0':
        field_std = 2./(1.+np.ma.exp(-relh_slope*data))-1.
        return field_std

    if (mx is None) or (mn is None):
        dlimits_dict = _data_limits_table()
        if field_name not in dlimits_dict:
            raise ValueError(
                'Field ' + field_name + ' unknown. '
                + 'Valid field names for standardizing are: '
                + 'H_ISO0, dBZ, ZDR, KDP and RhoHV')
        mx, mn = dlimits_dict[field_name]

    if field_name == 'KDP':
        data[data < -0.5] = -0.5
        data = 10.*np.ma.log10(data+0.6)
    elif field_name == 'RhoHV':
        # avoid infinite result
        data[data > 1.] = 1.
        data = 10.*np.ma.log10(1.0000000000001-data)

    mask = np.ma.getmaskarray(data)
    field_std = 2.*(data-mn)/(mx-mn)-1.
    field_std[data < mn] = -1.
    field_std[data > mx] = 1.
    field_std[mask] = np.ma.masked

    return field_std


def _destandardize(data, field_name, mx=None, mn=None, relh_slope=0.005):
    """
    destandardize the data

    Parameters
    ----------
    data : array
        Radar field.
    field_name : str
        Type of field (H_ISO0, dBZ, ZDR, KDP or RhoHV).
    mx, mn : floats or None, optional
        Data limits for array values.
    relh_slope : float
        The slope used to transform the relative height into a sigmoid
        function

    Returns
    -------
    field_std : dict
        destandardized radar data.

    """
    if field_name == 'H_ISO0':
        field_std = np.log(2./(1.+data)-1.)/(-relh_slope)
        return field_std

    if (mx is None) or (mn is None):
        dlimits_dict = _data_limits_table()
        if field_name not in dlimits_dict:
            raise ValueError(
                'Field ' + field_name + ' unknown. '
                + 'Valid field names for standardizing are: '
                + 'H_ISO0, dBZ, ZDR, KDP and RhoHV')
        mx, mn = dlimits_dict[field_name]

    if field_name == 'KDP':
        data = np.power(10., 0.1*data)-0.6
    elif field_name == 'RhoHV':
        data = 1.0000000000001-np.power(10., 0.1*data)

    mask = np.ma.getmaskarray(data)
    field_std = 0.5*(data+1.)*(mx-mn)+mn
    field_std[mask] = np.ma.masked

    return field_std


def _assign_to_class(fields_dict, mass_centers,
                     var_names=('dBZ', 'ZDR', 'KDP', 'RhoHV', 'H_ISO0'),
                     weights=np.array([1., 1., 1., 0.75, 0.5]),
                     t_vals=None):
    """
    Assigns an hydrometeor class to a radar range bin computing
    the distance between the radar variables an a centroid.

    Parameters
    ----------
    fields_dict : dict
        Dictionary containg the variables used for assigment normalized to
        [-1, 1] values
    mass_centers : matrix
        centroids normalized to [-1, 1] values (nclasses, nvariables)
    var_names : array of str
        Name of the variables
    weights : array
        optional. The weight given to each variable (nvariables)
    t_vals : array
        transformation values for the distance to centroids (nclasses)

    Returns
    -------
    hydroclass : int array
        the index corresponding to the assigned class
    entropy : float array
        the entropy
    t_dist : float matrix
        if entropy is computed, the transformed distances of each class
        (proxy for proportions of each hydrometeor) (nrays, nbins, nclasses)

    """
    # prepare data
    nrays = fields_dict[var_names[0]].shape[0]
    nbins = fields_dict[var_names[0]].shape[1]
    nclasses = mass_centers.shape[0]
    nvariables = mass_centers.shape[1]
    dtype = fields_dict[var_names[0]].dtype

    hydroclass = np.ma.empty((nrays, nbins), dtype=np.uint8)
    entropy = None
    t_dist = None
    if t_vals is not None:
        entropy = np.ma.empty((nrays, nbins), dtype=dtype)
        t_dist = np.ma.masked_all((nrays, nbins, nclasses), dtype=dtype)

    for ray in range(nrays):
        data = []
        for var_name in var_names:
            data.append(fields_dict[var_name][ray, :])
        data = np.ma.array(data, dtype=dtype)
        weights_mat = np.broadcast_to(
            weights.reshape(nvariables, 1), (nvariables, nbins))
        dist = np.ma.zeros((nclasses, nbins), dtype=dtype)

        # compute distance: masked entries will not contribute to the distance
        mask = np.ma.getmaskarray(fields_dict[var_names[0]][ray, :])
        for i in range(nclasses):
            centroids_class = mass_centers[i, :]
            centroids_class = np.broadcast_to(
                centroids_class.reshape(nvariables, 1), (nvariables, nbins))
            dist_ray = np.ma.sqrt(np.ma.sum(
                ((centroids_class-data)**2.)*weights_mat, axis=0))
            dist_ray[mask] = np.ma.masked
            dist[i, :] = dist_ray

        # Get hydrometeor class
        class_vec = dist.argsort(axis=0, fill_value=10e40)
        hydroclass_ray = (class_vec[0, :]+2).astype(np.uint8)
        hydroclass_ray[mask] = 1
        hydroclass[ray, :] = hydroclass_ray

        if t_vals is None:
            continue

        # Transform the distance using the coefficient of the dominant class
        t_vals_ray = np.ma.masked_where(mask, t_vals[class_vec[0, :]])
        t_vals_ray = ma_broadcast_to(
            t_vals_ray.reshape(1, nbins), (nclasses, nbins))
        t_dist_ray = np.ma.exp(-t_vals_ray*dist)

        # set transformed distances to a value between 0 and 1
        dist_total = np.ma.sum(t_dist_ray, axis=0)
        dist_total = ma_broadcast_to(
            dist_total.reshape(1, nbins), (nclasses, nbins))
        t_dist_ray /= dist_total

        # Compute entropy
        entropy_ray = -np.ma.sum(
            t_dist_ray*np.ma.log(t_dist_ray)/np.ma.log(nclasses), axis=0)
        entropy_ray[mask] = np.ma.masked
        entropy[ray, :] = entropy_ray

        t_dist[ray, :, :] = np.ma.transpose(t_dist_ray)

    if t_vals is not None:
        t_dist *= 100.

    return hydroclass, entropy, t_dist


def _assign_to_class_scan(fields_dict, mass_centers,
                          var_names=('dBZ', 'ZDR', 'KDP', 'RhoHV', 'H_ISO0'),
                          weights=np.array([1., 1., 1., 0.75, 0.5]),
                          t_vals=None):
    """
    assigns an hydrometeor class to a radar range bin computing
    the distance between the radar variables an a centroid.
    Computes the entire radar volume at once

    Parameters
    ----------
    fields_dict : dict
        Dictionary containg the variables used for assigment normalized to
        [-1, 1] values
    mass_centers : matrix
        centroids normalized to [-1, 1] values
    var_names : array of str
        Name of the variables
    weights : array
        optional. The weight given to each variable
    t_vals : matrix
        transformation values for the distance to centroids
        (nclasses, nvariables)

    Returns
    -------
    hydroclass : int array
        the index corresponding to the assigned class
    entropy : float array
        the entropy
    t_dist : float matrix
        if entropy is computed, the transformed distances of each class
        (proxy for proportions of each hydrometeor) (nrays, nbins, nclasses)

    """
    # prepare data
    nrays = fields_dict[var_names[0]].shape[0]
    nbins = fields_dict[var_names[0]].shape[1]
    nclasses = mass_centers.shape[0]
    nvariables = mass_centers.shape[1]
    dtype = fields_dict[var_names[0]].dtype

    data = []
    for var_name in var_names:
        data.append(fields_dict[var_name])
    data = np.ma.array(data, dtype=dtype)
    weights_mat = np.broadcast_to(
        weights.reshape(nvariables, 1, 1), (nvariables, nrays, nbins))

    # compute distance: masked entries will not contribute to the distance
    mask = np.ma.getmaskarray(fields_dict[var_names[0]])
    dist = np.ma.zeros((nrays, nbins, nclasses), dtype=dtype)
    t_dist = None
    entropy = None
    for i in range(nclasses):
        centroids_class = mass_centers[i, :]
        centroids_class = np.broadcast_to(
            centroids_class.reshape(nvariables, 1, 1),
            (nvariables, nrays, nbins))
        dist_aux = np.ma.sqrt(np.ma.sum(
            ((centroids_class-data)**2.)*weights_mat, axis=0))
        dist_aux[mask] = np.ma.masked
        dist[:, :, i] = dist_aux

    del data
    del weights_mat

    # Get hydrometeor class
    class_vec = dist.argsort(axis=-1, fill_value=10e40)
    hydroclass = np.ma.asarray(class_vec[:, :, 0]+2, dtype=np.uint8)
    hydroclass[mask] = 1

    if t_vals is not None:
        # Transform the distance using the coefficient of the dominant class
        t_vals_aux = np.ma.masked_where(mask, t_vals[class_vec[:, :, 0]])
        t_vals_aux = ma_broadcast_to(
            t_vals_aux.reshape(nrays, nbins, 1), (nrays, nbins, nclasses))
        t_dist = np.ma.exp(-t_vals_aux*dist)
        del t_vals_aux

        # set distance to a value between 0 and 1
        dist_total = np.ma.sum(t_dist, axis=-1)
        dist_total = ma_broadcast_to(
            dist_total.reshape(nrays, nbins, 1), (nrays, nbins, nclasses))
        t_dist /= dist_total
        del dist_total

        # compute entroy
        entropy = -np.ma.sum(
            t_dist*np.ma.log(t_dist)/np.ma.log(nclasses), axis=-1)
        entropy[mask] = np.ma.masked

        t_dist *= 100.

    return hydroclass, entropy, t_dist


def _compute_coeff_transform(mass_centers,
                             weights=np.array([1., 1., 1., 0.75, 0.5]),
                             value=50.):
    """
    get the transformation coefficients

    Parameters
    ----------
    mass_centers : ndarray 2D
        The centroids for each class and variable (nclasses, nvariables)
    weights : array
        optional. The weight given to each variable (nvariables)
    value : float
        parameter controlling the rate of decay of the distance transformation

    Returns
    -------
    t_vals : ndarray 1D
        The coefficients used to transform the distances to each centroid for
        each class (nclasses)

    """
    nclasses, nvariables = np.shape(mass_centers)
    t_vals = np.empty((nclasses, nclasses), dtype=mass_centers.dtype)
    for i in range(nclasses):
        weights_mat = np.broadcast_to(
            weights.reshape(1, nvariables), (nclasses, nvariables))
        centroids_class = mass_centers[i, :]
        centroids_class = np.broadcast_to(
            centroids_class.reshape(1, nvariables), (nclasses, nvariables))
        t_vals[i, :] = np.sqrt(
            np.sum(weights_mat*np.power(
                np.abs(centroids_class-mass_centers), 2.), axis=1))

    # pick the second lowest value (the first is 0)
    t_vals = np.sort(t_vals, axis=-1)[:, 1]
    t_vals = np.log(value)/t_vals

    return t_vals


def _get_mass_centers(freq):
    """
    Get mass centers for a particular frequency.

    Parameters
    ----------
    freq : float
        Radar frequency [Hz].

    Returns
    -------
    mass_centers : ndarray 2D
        The centroids for each variable and hydrometeor class in (nclasses,
        nvariables).

    """
    mass_centers_dict = _mass_centers_table()

    freq_band = get_freq_band(freq)
    if (freq_band is not None) and (freq_band in mass_centers_dict):
        return mass_centers_dict[freq_band]

    if freq < 4e9:
        freq_band_aux = 'C'
    elif freq > 12e9:
        freq_band_aux = 'X'

    mass_centers = mass_centers_dict[freq_band_aux]
    warn('Radar frequency out of range. ' +
         'Centroids only valid for C or X band. ' +
         freq_band_aux + ' band centroids will be applied')

    return mass_centers


def _mass_centers_table():
    """
    Defines the mass centers look up table for each frequency band.

    Returns
    -------
    mass_centers_dict : dict
        A dictionary with the mass centers for each frequency band.

    """
    nclasses = 9
    nvariables = 5
    mass_centers = np.zeros((nclasses, nvariables))

    mass_centers_dict = dict()
    # C-band centroids derived for MeteoSwiss Albis radar
    #                       Zh        ZDR     kdp   RhoHV    delta_Z
    mass_centers[0, :] = [13.5829, 0.4063, 0.0497, 0.9868, 1330.3]  # DS
    mass_centers[1, :] = [02.8453, 0.2457, 0.0000, 0.9798, 0653.8]  # CR
    mass_centers[2, :] = [07.6597, 0.2180, 0.0019, 0.9799, -1426.5]  # LR
    mass_centers[3, :] = [31.6815, 0.3926, 0.0828, 0.9978, 0535.3]  # GR
    mass_centers[4, :] = [39.4703, 1.0734, 0.4919, 0.9876, -1036.3]  # RN
    mass_centers[5, :] = [04.8267, -0.5690, 0.0000, 0.9691, 0869.8]  # VI
    mass_centers[6, :] = [30.8613, 0.9819, 0.1998, 0.9845, -0066.1]  # WS
    mass_centers[7, :] = [52.3969, 2.1094, 2.4675, 0.9730, -1550.2]  # MH
    mass_centers[8, :] = [50.6186, -0.0649, 0.0946, 0.9904, 1179.9]  # IH/HDG

    mass_centers_dict.update({'C': mass_centers})

    # X-band centroids derived for MeteoSwiss DX50 radar
    #                       Zh        ZDR     kdp    RhoHV   delta_Z
    mass_centers[0, :] = [19.0770, 0.4139, 0.0099, 0.9841, 1061.7]  # DS
    mass_centers[1, :] = [03.9877, 0.5040, 0.0000, 0.9642, 0856.6]  # CR
    mass_centers[2, :] = [20.7982, 0.3177, 0.0004, 0.9858, -1375.1]  # LR
    mass_centers[3, :] = [34.7124, -0.3748, 0.0988, 0.9828, 1224.2]  # GR
    mass_centers[4, :] = [33.0134, 0.6614, 0.0819, 0.9802, -1169.8]  # RN
    mass_centers[5, :] = [08.2610, -0.4681, 0.0000, 0.9722, 1100.7]  # VI
    mass_centers[6, :] = [35.1801, 1.2830, 0.1322, 0.9162, -0159.8]  # WS
    mass_centers[7, :] = [52.4539, 2.3714, 1.1120, 0.9382, -1618.5]  # MH
    mass_centers[8, :] = [44.2216, -0.3419, 0.0687, 0.9683, 1272.7]  # IH/HDG

    mass_centers_dict.update({'X': mass_centers})

    # S-band centroids: Dummy centroids derived for MeteoSwiss C-band Albis
    # radar. To be substituted for real S-band ones
    #                       Zh        ZDR     kdp   RhoHV    delta_Z
    mass_centers[0, :] = [13.5829, 0.4063, 0.0497, 0.9868, 1330.3]  # DS
    mass_centers[1, :] = [02.8453, 0.2457, 0.0000, 0.9798, 0653.8]  # CR
    mass_centers[2, :] = [07.6597, 0.2180, 0.0019, 0.9799, -1426.5]  # LR
    mass_centers[3, :] = [31.6815, 0.3926, 0.0828, 0.9978, 0535.3]  # GR
    mass_centers[4, :] = [39.4703, 1.0734, 0.4919, 0.9876, -1036.3]  # RN
    mass_centers[5, :] = [04.8267, -0.5690, 0.0000, 0.9691, 0869.8]  # VI
    mass_centers[6, :] = [30.8613, 0.9819, 0.1998, 0.9845, -0066.1]  # WS
    mass_centers[7, :] = [52.3969, 2.1094, 2.4675, 0.9730, -1550.2]  # MH
    mass_centers[8, :] = [50.6186, -0.0649, 0.0946, 0.9904, 1179.9]  # IH/HDG

    mass_centers_dict.update({'S': mass_centers})

    return mass_centers_dict


def _data_limits_table():
    """
    Defines the data limits used in the standardization.

    Returns
    -------
    dlimits_dict : dict
        A dictionary with the limits for each variable.

    """
    dlimits_dict = dict()
    dlimits_dict.update({'dBZ': (60., -10.)})
    dlimits_dict.update({'ZDR': (5., -1.5)})
    dlimits_dict.update({'KDP': (7., -10.)})
    dlimits_dict.update({'RhoHV': (5.23, -50.)})
    dlimits_dict.update({'H_ISO0': (5000., -5000.)})

    return dlimits_dict


def _data_limits_centroids_table():
    """
    Defines the data limits used in the computation of the centroids.

    Returns
    -------
    dlimits_dict : dict
        A dictionary with the limits for each variable.

    """
    dlimits_dict = dict()
    dlimits_dict.update({'dBZ': (60., -9.)})
    dlimits_dict.update({'ZDR': (5., -1.5)})
    dlimits_dict.update({'KDP': (5., -0.5)})
    dlimits_dict.update({'RhoHV': (1., 0.7)})
    dlimits_dict.update({'H_ISO0': (2000., -2000.)})

    return dlimits_dict


def _bell_function_table():
    """
    Defines the bell-shaped functions of the polarimetric variables for each
    hydrometeor type

    Returns
    -------
    bell_function_dict : dict of dicts
        A dictionary with the parameters defining the bell-shaped function for
        each frequency, variable and hydrometeor type

    """
    bell_function_dict = dict()

    # C-band m, a, b
    dBZ_dict = {
        'CR': (-2.8, 12., 5.),
        'AG': (17., 18.1, 10.),
        'LR': (1.75, 29., 10.),
        'RN': (39., 19., 10.),
        'RP': (37., 9.2, 0.8),
        'VI': (-1., 11., 5.),
        'WS': (24., 21.3, 10.),
        'MH': (58.18, 8., 10.),
        'IH/HDG': (48.8, 8., 10.)}

    ZDR_dict = {
        'CR': (2.9, 2.7, 10.),
        'AG': (1., 1.1, 7.),
        'LR': (0.46, 0.46, 5.),
        'RN': (2.3, 2.2, 9.),
        'RP': (0.9, 0.9, 6.),
        'VI': (-0.9, 0.9, 10),
        'WS': (1.3, 0.9, 10.),
        'MH': (2.19, 1.5, 10.),
        'IH/HDG': (0.36, 0.5, 10.)}

    KDP_dict = {
        'CR': (0.08, 0.08, 6.),
        'AG': (-0.008, 0.3, 1.),
        'LR': (0.03, 0.03, 2.),
        'RN': (5.5, 5.5, 10.),
        'RP': (0.1, 0.08, 3.),
        'VI': (-0.75, 0.75, 30.),
        'WS': (0.25, 0.43, 6.),
        'MH': (1.08, 2., 6.),
        'IH/HDG': (0.07, 0.15, 6.)}

    RhoHV_dict = {
        'CR': (0.98, 0.025, 3.),
        'AG': (0.93, 0.07, 3.),
        'LR': (1., 0.018, 3.),
        'RN': (1., 0.025, 3.),
        'RP': (1., 0.025, 1.),
        'VI': (0.975, 0.022, 3.),
        'WS': (0.8, 0.10, 10.),
        'MH': (0.95, 0.05, 3.),
        'IH/HDG': (0.99, 0.05, 3.)}

    vars_dict = {
        'dBZ': dBZ_dict,
        'ZDR': ZDR_dict,
        'KDP': KDP_dict,
        'RhoHV': RhoHV_dict}

    bell_function_dict.update({'C': vars_dict})

    # X-band m, a, b
    dBZ_dict = {
        'CR': (-3, 12., 5.),
        'AG': (16., 17., 10.),
        'LR': (2., 29., 10.),
        'RN': (42., 17., 10.),
        'RP': (34., 10., 0.8),
        'VI': (3.5, 14., 5.),
        'WS': (30., 20., 10.),
        'MH': (53.37, 8., 10.),
        'IH/HDG': (45.5, 8., 10.)}

    ZDR_dict = {
        'CR': (3.2, 2.6, 10.),
        'AG': (0.7, 0.7, 7.),
        'LR': (0.5, 0.5, 5.),
        'RN': (2.7, 2.8, 9.),
        'RP': (0.3, 1., 6.),
        'VI': (-0.8, 1.3, 10.),
        'WS': (1.3, 0.9, 10.),
        'MH': (2.2, 1.4, 10.),
        'IH/HDG': (2.6, 0.5, 10.)}

    KDP_dict = {
        'CR': (0.15, 0.15, 6.),
        'AG': (0.2, 0.2, 1.),
        'LR': (0.18, 0.18, 2.),
        'RN': (12.6, 12.9, 10.),
        'RP': (0.7, 2.1, 3.),
        'VI': (-0.1, 0.08, 30.),
        'WS': (1., 1., 6.),
        'MH': (1.37, 2., 6.),
        'IH/HDG': (0.1, 0.15, 6.)}

    RhoHV_dict = {
        'CR': (0.985, 0.015, 3.),
        'AG': (0.989, 0.011, 3.),
        'LR': (0.992, 0.007, 3.),
        'RN': (0.99, 0.01, 3.),
        'RP': (0.993, 0.007, 1.),
        'VI': (0.965, 0.035, 3.),
        'WS': (0.835, 0.135, 10.),
        'MH': (0.96, 0.05, 3.),
        'IH/HDG': (0.97, 0.05, 3.)}

    vars_dict = {
        'dBZ': dBZ_dict,
        'ZDR': ZDR_dict,
        'KDP': KDP_dict,
        'RhoHV': RhoHV_dict}

    bell_function_dict.update({'X': vars_dict})

    # S-band m, a, b: WS, MH, IH to be reviewed
    dBZ_dict = {
        'CR': (-3, 12., 5.),
        'AG': (17., 17., 10.),
        'LR': (-3., 24., 10.),
        'RN': (41.5, 15.5, 10.),
        'RP': (35., 10., 0.8),
        'VI': (3., 14., 5.),
        'WS': (35., 20., 10.),
        'MH': (55., 8., 10.),
        'IH/HDG': (45., 8., 10.)}

    ZDR_dict = {
        'CR': (2.9, 2.9, 10.),
        'AG': (0.6, 0.6, 7.),
        'LR': (0.35, 0.35, 5.),
        'RN': (2.6, 2.5, 9.),
        'RP': (0.3, 0.8, 6.),
        'VI': (-0.8, 1.3, 10.),
        'WS': (1.5, 0.9, 10.),
        'MH': (2., 1.5, 10.),
        'IH/HDG': (0., 0.5, 10.)}

    KDP_dict = {
        'CR': (0.045, 0.045, 6.),
        'AG': (0.04, 0.04, 1.),
        'LR': (0.01, 0.01, 2.),
        'RN': (3.7, 3.7, 10.),
        'RP': (0.2, 0.6, 3.),
        'VI': (-0.02, 0.02, 30.),
        'WS': (0.1, 0.4, 6.),
        'MH': (0.5, 2., 6.),
        'IH/HDG': (0.05, 0.15, 6.)}

    RhoHV_dict = {
        'CR': (0.99, 0.01, 3.),
        'AG': (0.989, 0.011, 3.),
        'LR': (0.995, 0.005, 3.),
        'RN': (0.99, 0.01, 3.),
        'RP': (0.995, 0.005, 1.),
        'VI': (0.965, 0.035, 3.),
        'WS': (0.85, 0.1, 10.),
        'MH': (0.95, 0.05, 3.),
        'IH/HDG': (0.98, 0.05, 3.)}

    vars_dict = {
        'dBZ': dBZ_dict,
        'ZDR': ZDR_dict,
        'KDP': KDP_dict,
        'RhoHV': RhoHV_dict}

    bell_function_dict.update({'S': vars_dict})

    return bell_function_dict


def _trapezoidal_function_table():
    """
    Defines the trapezoidal functions of the height respect to iso-0 for each
    hydrometeor type

    Returns
    -------
    trapezoidal_function_dict : dict
        A dictionary with the parameters defining the trapezoidal function for
        each hydrometeor type

    """
    # v1, v2, v3, v4
    return {
        'CR': (0., 1000., 2200., 2500.),
        'AG': (0., 500., 2000., 2500.),
        'LR': (-2500., -2200., -300., 0.),
        'RN': (-2500., -2200., -300., 0.),
        'RP': (0., 500., 2000., 2200.),
        'VI': (0., 1000., 2200., 2500.),
        'WS': (-500., -300., 300., 500.),
        'MH': (-2500., -2200., -300., 0.),
        'IH/HDG': (0., 500., 2000., 2500.)}



"""
pyart.aux_io.edge_necdf
=======================

Utilities for reading EDGE NetCDF files.

.. autosummary::
    :toctree: generated/

    read_edge_netcdf

"""

import datetime

import numpy as np
import netCDF4

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar


def read_edge_netcdf(filename, **kwargs):
    """
    Read a EDGE NetCDF file.

    Parameters
    ----------
    filename : str
        Name of EDGE NetCDF file to read data from.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('edge_netcdf')

    # Open netCDF4 file
    dset = netCDF4.Dataset(filename)
    nrays = len(dset.dimensions['Azimuth'])
    nbins = len(dset.dimensions['Gate'])

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    latitude['data'] = np.array([dset.Latitude], 'float64')
    longitude['data'] = np.array([dset.Longitude], 'float64')
    altitude['data'] = np.array([dset.Height], 'float64')

    # metadata
    metadata = filemetadata('metadata')
    metadata_mapping = {
        'vcp-value': 'vcp',
        'radarName-value': 'radar_name',
        'ConversionPlugin': 'conversion_software',
    }
    for netcdf_attr, metadata_key in metadata_mapping.items():
        if netcdf_attr in dset.ncattrs():
            metadata[metadata_key] = dset.getncattr(netcdf_attr)

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_start_ray_index['data'] = np.array([0], dtype='int32')
    sweep_end_ray_index['data'] = np.array([nrays-1], dtype='int32')

    # sweep number
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.array([0], dtype='int32')

    # sweep_type
    scan_type = 'ppi'

    # sweep_mode, fixed_angle
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    sweep_mode['data'] = np.array(1 * ['azimuth_surveillance'])
    fixed_angle['data'] = np.array([dset.Elevation], dtype='float32')

    # time
    time = filemetadata('time')
    start_time = datetime.datetime.utcfromtimestamp(dset.Time)
    time['units'] = make_time_unit_str(start_time)
    time['data'] = np.zeros((nrays, ), dtype='float64')

    # range
    _range = filemetadata('range')
    step = float(dset.getncattr('MaximumRange-value')) / nbins * 1000.
    _range['data'] = (np.arange(nbins, dtype='float32') * step + step / 2)
    _range['meters_to_center_of_first_gate'] = step / 2.
    _range['meters_between_gates'] = step

    # elevation
    elevation = filemetadata('elevation')
    elevation_angle = dset.Elevation
    elevation['data'] = np.ones((nrays, ), dtype='float32') * elevation_angle

    # azimuth
    azimuth = filemetadata('azimuth')
    azimuth['data'] = dset.variables['Azimuth'][:]

    # fields
    field_name = dset.TypeName

    field_data = np.ma.array(dset.variables[field_name][:])
    if 'MissingData' in dset.ncattrs():
        field_data[field_data == dset.MissingData] = np.ma.masked
    if 'RangeFolded' in dset.ncattrs():
        field_data[field_data == dset.RangeFolded] = np.ma.masked

    fields = {field_name: filemetadata(field_name)}
    fields[field_name]['data'] = field_data
    fields[field_name]['units'] = dset.variables[field_name].Units
    fields[field_name]['_FillValue'] = get_fillvalue()

    # instrument_parameters
    instrument_parameters = {}

    if 'PRF-value' in dset.ncattrs():
        dic = filemetadata('prt')
        prt = 1. / float(dset.getncattr('PRF-value'))
        dic['data'] = np.ones((nrays, ), dtype='float32') * prt
        instrument_parameters['prt'] = dic

    if 'PulseWidth-value' in dset.ncattrs():
        dic = filemetadata('pulse_width')
        pulse_width = dset.getncattr('PulseWidth-value') * 1.e-6
        dic['data'] = np.ones((nrays, ), dtype='float32') * pulse_width
        instrument_parameters['pulse_width'] = dic

    if 'NyquistVelocity-value' in dset.ncattrs():
        dic = filemetadata('nyquist_velocity')
        nyquist_velocity = float(dset.getncattr('NyquistVelocity-value'))
        dic['data'] = np.ones((nrays, ), dtype='float32') * nyquist_velocity
        instrument_parameters['nyquist_velocity'] = dic

    if 'Beamwidth' in dset.variables:
        dic = filemetadata('radar_beam_width_h')
        dic['data'] = dset.variables['Beamwidth'][:]
        instrument_parameters['radar_beam_width_h'] = dic

    dset.close()

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)



"""
======================================
Map a single radar to a Cartesian grid
======================================

Map the reflectivity field of a single radar from Antenna coordinates to a
Cartesian grid.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import numpy as np
import matplotlib.pyplot as plt
import pyart

# read in the data
RADAR_FILE = '110635.mdv'
radar = pyart.io.read_mdv(RADAR_FILE)

# mask out last 10 gates of each ray, this removes the "ring" around th radar.
radar.fields['reflectivity']['data'][:, -10:] = np.ma.masked

# exclude masked gates from the gridding
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_transition()
gatefilter.exclude_masked('reflectivity')

# perform Cartesian mapping, limit to the reflectivity field.
grid = pyart.map.grid_from_radars(
    (radar,), gatefilters=(gatefilter, ),
    grid_shape=(1, 241, 241),
    grid_limits=((2000, 2000), (-123000.0, 123000.0), (-123000.0, 123000.0)),
    fields=['reflectivity'])

# create the plot
fig = plt.figure()
ax = fig.add_subplot(111)
ax.imshow(grid.fields['reflectivity']['data'][0], origin='lower')
plt.show()



"""
pyart.retrieve.wind
=========================================

Functions for wind estimation

.. autosummary::
    :toctree: generated/

    est_wind_vel
    est_vertical_windshear
    est_wind_profile
    _wind_coeff
    _vad
    _vel_variance

"""

from copy import deepcopy

import numpy as np

from ..config import get_metadata, get_field_name


def est_wind_vel(radar, vert_proj=False, vel_field=None, wind_field=None):
    """
    Estimates wind velocity. Projects the radial wind component to the
    horizontal or vertical of the azimuth plane. It assumes that the
    orthogonal component is negligible.

    The horizontal wind component is given by:
        v = v_r*cos(el)-v_el*sin(el)+v_az
    where:
        v_r is the radial wind component (measured by the radar)
        v_el is the perpendicular wind component in the azimuth plane.
        v_az is the horizontal component perpendicular to the radial
        direction and the azimuth plane
        el is the elevation

    The horizontal wind component in the azimuth plane is given by:
        v_h = v_r*cos(el) - v_el*sin(el)
    which since we do not know v_el we assume:
        v_h ~ v_r*cos(el)
    This assumption holds for small elevation angles

    The vertical wind component in the azimuth plane is given by:
        v_h = v_r*sin(el) - v_el*cos(el)
    which since we do not know v_el we assume:
        v_h ~ v_r*sin(el)
    This assumption holds for angles close to 90 deg

    Parameters
    ----------
    radar : Radar
        Radar object
    vert_proj : Boolean
        If true estimates the vertical projection, otherwise the horizontal
    vel_field : str
        name of the velocity field
    wind_field : str
        name of the velocity field

    Returns
    -------
    wind : dict
        Field dictionary containing the estimated wind velocity

    """

    # parse the field parameters
    if vel_field is None:
        vel_field = get_field_name('velocity')
    if wind_field is None:
        wind_field = get_field_name('azimuthal_horizontal_wind_component')
        if vert_proj:
            wind_field = get_field_name('vertical_wind_component')

    radar.check_field_exists(vel_field)

    vel = radar.fields[vel_field]['data']
    ele = deepcopy(radar.elevation['data'])
    ele[ele > 90.] = 180.-ele[ele > 90.]
    ele = np.broadcast_to(ele.reshape(radar.nrays, 1),
                          (radar.nrays, radar.ngates))
    if vert_proj:
        wind_data = vel * np.sin(ele*np.pi/180.)
    else:
        wind_data = vel * np.cos(ele*np.pi/180.)

    wind = get_metadata(wind_field)
    wind['data'] = wind_data

    return wind


def est_vertical_windshear(radar, az_tol=0.5, wind_field=None,
                           windshear_field=None):
    """
    Estimates wind shear.

    Parameters
    ----------
    radar : Radar
        Radar object
    az_tol : float
        azimuth tolerance to consider gate on top of selected one
    wind_field : str
        name of the horizontal wind velocity field
    windshear_field : str
        name of the vertical wind shear field

    Returns
    -------
    windshear : dict
        Field dictionary containing the wind shear field

    """

    # parse the field parameters
    if wind_field is None:
        wind_field = get_field_name('azimuthal_horizontal_wind_component')
    if windshear_field is None:
        windshear_field = get_field_name('vertical_wind_shear')

    radar.check_field_exists(wind_field)
    wind = radar.fields[wind_field]['data']

    # compute ground range
    ele = deepcopy(radar.elevation['data'])
    ele[ele > 90.] = 180.-ele[ele > 90.]
    ele = np.broadcast_to(ele.reshape(radar.nrays, 1),
                          (radar.nrays, radar.ngates))

    rng_mat = np.broadcast_to(radar.range['data'].reshape(1, radar.ngates),
                              (radar.nrays, radar.ngates))
    rng_ground = rng_mat*np.cos(ele*np.pi/180.)

    # initialize output
    windshear_data = np.ma.empty((radar.nrays, radar.ngates), dtype=float)
    windshear_data[:] = np.ma.masked

    azi_vec = radar.azimuth['data']
    ele_vec = radar.elevation['data']
    alt_mat = radar.gate_altitude['data']
    mask = np.ma.getmaskarray(wind)
    for ray in range(radar.nrays):
        # look for the elevation on top of the current ray
        ind_rays_top = np.where(
            np.logical_and(
                ele_vec > ele_vec[ray],
                np.abs(azi_vec-azi_vec[ray]) < az_tol))[0]
        if ind_rays_top.size == 0:
            continue
        ind_rays_top = np.where(ele_vec == np.min(ele_vec[ind_rays_top]))[0]
        ele_nearest = ele_vec[ind_rays_top[0]]
        azi_top = azi_vec[ind_rays_top]
        azi_nearest = azi_top[np.argmin(np.abs(azi_top-azi_vec[ray]))]
        ind_ray = np.where(
            np.logical_and(
                ele_vec == ele_nearest, azi_vec == azi_nearest))[0][0]

        for rng in range(radar.ngates):
            if mask[ray, rng]:
                continue
            # look for the nearest gate on top of selected gate
            ind_rng = np.argmin(
                np.abs(rng_ground[ind_ray, :]-rng_ground[ray, rng]))
            if mask[ind_ray, ind_rng]:
                continue
            # interpolate the two closest gates on top of the one
            # examined
            if rng_ground[ind_ray, ind_rng] < rng_ground[ray, rng]:
                if ind_rng+1 >= radar.ngates:
                    continue
                if mask[ind_ray, ind_rng+1]:
                    continue
                rng_ground_near = rng_ground[ind_ray, ind_rng:ind_rng+2]
                wind_near = wind[ind_ray, ind_rng:ind_rng+2]
                alt_near = alt_mat[ind_ray, ind_rng:ind_rng+2]
            else:
                if ind_rng-1 < 0:
                    continue
                if mask[ind_ray, ind_rng-1]:
                    continue
                rng_ground_near = rng_ground[ind_ray, ind_rng-1:ind_rng+1]
                wind_near = wind[ind_ray, ind_rng-1:ind_rng+1]
                alt_near = alt_mat[ind_ray, ind_rng-1:ind_rng+1]

            wind_top = np.interp(
                rng_ground[ray, rng], rng_ground_near, wind_near)
            gate_altitude_top = np.interp(
                rng_ground[ray, rng], rng_ground_near, alt_near)
            # compute wind shear
            windshear_data[ray, rng] = (
                1000.*(wind_top-wind[ray, rng]) /
                (gate_altitude_top-alt_mat[ray, rng]))

    windshear = get_metadata(windshear_field)
    windshear['data'] = windshear_data

    return windshear


def est_wind_profile(radar, npoints_min=6, azi_spacing_max=45.,
                     vel_diff_max=10., sign=1, rad_vel_field=None,
                     u_vel_field=None, v_vel_field=None, w_vel_field=None,
                     vel_est_field=None, vel_std_field=None,
                     vel_diff_field=None):
    """
    Estimates the vertical wind profile using VAD techniques

    Parameters
    ----------
    radar : Radar
        Radar object
    npoints_min : int
        Minimum number of points in the VAD to retrieve wind components.
        0 will retrieve them regardless
    azi_spacing_max : float
        Maximum spacing between valid gates in the VAD to retrieve wind
        components. 0 will retrieve them regardless.
    vel_diff_max : float
        Maximum velocity difference allowed between retrieved and measured
        radial velocity at each range gate. Gates exceeding this threshold
        will be removed and VAD will be recomputed. If -1 there will not be
        a second pass.
    sign : int, optional
        Sign convention which the radial velocities in the volume created
        from the sounding data will will. This should match the convention
        used in the radar data. A value of 1 represents when positive values
        velocities are towards the radar, -1 represents when negative
        velocities are towards the radar.
    rad_vel_field : str
        name of the measured radial velocity field
    u_vel_field, v_vel_field, w_vel_field : str
        names of the 3 wind components fields
    vel_est_field : str
        name of the retrieved radial Doppler velocity field
    vel_std_field : str
        name of the standard deviation of the velocity retrieval field
    vel_diff_field : str
        name of the diference between retrieved and measured radial velocity
        field

    Returns
    -------
    wind : dict
        Field dictionary containing the estimated wind velocity

    """
    # parse the field parameters
    if rad_vel_field is None:
        rad_vel_field = get_field_name('velocity')
    if u_vel_field is None:
        u_vel_field = get_field_name('eastward_wind_component')
    if v_vel_field is None:
        v_vel_field = get_field_name('northward_wind_component')
    if w_vel_field is None:
        w_vel_field = get_field_name('vertical_wind_component')
    if vel_est_field is None:
        vel_est_field = get_field_name('retrieved_velocity')
    if vel_std_field is None:
        vel_std_field = get_field_name('retrieved_velocity_std')
    if vel_diff_field is None:
        vel_diff_field = get_field_name('velocity_difference')

    radar.check_field_exists(rad_vel_field)
    vel = deepcopy(radar.fields[rad_vel_field]['data'])

    # Compute wind coefficients
    u_coeff, v_coeff, w_coeff = _wind_coeff(radar)

    # first guess VAD
    u_vel, v_vel, w_vel, vel_est = _vad(
        radar, u_coeff, v_coeff, w_coeff, vel, npoints_min=npoints_min,
        azi_spacing_max=azi_spacing_max, sign=sign)

    # Remove gates where velocity difference exceeds threshold and recompute
    # VAD if applicable
    if vel_diff_max > -1:
        vel = np.ma.masked_where(np.ma.abs(vel-vel_est) > vel_diff_max, vel)

        # Final VAD
        u_vel, v_vel, w_vel, vel_est = _vad(
            radar, u_coeff, v_coeff, w_coeff, vel, npoints_min=npoints_min,
            azi_spacing_max=azi_spacing_max)

    # Compute velocity estimation variance
    vel_std, vel_diff = _vel_std(radar, vel, vel_est)

    # prepare output
    u_vel_dict = get_metadata(u_vel_field)
    u_vel_dict['data'] = u_vel

    v_vel_dict = get_metadata(v_vel_field)
    v_vel_dict['data'] = v_vel

    w_vel_dict = get_metadata(w_vel_field)
    w_vel_dict['data'] = w_vel

    vel_est_dict = get_metadata(vel_est_field)
    vel_est_dict['data'] = vel_est

    vel_std_dict = get_metadata(vel_std_field)
    vel_std_dict['data'] = vel_std

    vel_diff_dict = get_metadata(vel_diff_field)
    vel_diff_dict['data'] = vel_diff

    return (u_vel_dict, v_vel_dict, w_vel_dict, vel_est_dict, vel_std_dict,
            vel_diff_dict)


def _wind_coeff(radar):
    """
    Computes the coefficients to transform 3-D wind vectors into radial
    velocity at each range gate

    Parameters
    ----------
    radar : Radar
        Radar object

    Returns
    -------
    u_coeff, v_coeff, w_coeff : 2D float arrays
        The coefficients for each wind component

    """
    cos_ele = np.cos(radar.elevation['data']*np.pi/180.)
    sin_ele = np.sin(radar.elevation['data']*np.pi/180.)
    cos_azi = np.cos(radar.azimuth['data']*np.pi/180.)
    sin_azi = np.sin(radar.azimuth['data']*np.pi/180.)

    u_coeff = sin_azi*cos_ele
    v_coeff = cos_azi*cos_ele
    w_coeff = sin_ele

    u_coeff = np.reshape(
        np.tile(u_coeff, radar.ngates), (radar.ngates, radar.nrays)).T
    v_coeff = np.reshape(
        np.tile(v_coeff, radar.ngates), (radar.ngates, radar.nrays)).T
    w_coeff = np.reshape(
        np.tile(w_coeff, radar.ngates), (radar.ngates, radar.nrays)).T

    return u_coeff, v_coeff, w_coeff


def _vad(radar, u_coeff, v_coeff, w_coeff, vel, npoints_min=6,
         azi_spacing_max=45., sign=1):
    """
    Estimates wind components using VAD techniques

    Parameters
    ----------
    radar : Radar
        Radar object
    u_coeff, v_coeff, w_coeff : 2D float arrays
        the coefficients to transform 3D winds into radial velocity
    vel : 2D float array
        The measured radial velocity field
    npoints_min : int
        Minimum number of points in the VAD to retrieve wind components.
        0 will retrieve them regardless
    azi_spacing_max : float
        Maximum spacing between valid gates in the VAD to retrieve wind
        components. 0 will retrieve them regardless.
    sign : int, optional
        Sign convention which the radial velocities in the volume created
        from the sounding data will will. This should match the convention
        used in the radar data. A value of 1 represents when positive values
        velocities are towards the radar, -1 represents when negative
        velocities are towards the radar.

    Returns
    -------
    u_vel, v_vel, w_vel : 2D float arrays
        The 3 estimated wind components at each range gate
    vel_est : 2D float array
        The estimated radial velocity at each range gate

    """
    vel_aux = deepcopy(vel)
    if sign == 1:
        vel_aux = -vel_aux

    # prepare wind matrices
    u_vel = np.ma.empty((radar.nrays, radar.ngates), dtype=float)
    u_vel[:] = np.ma.masked
    v_vel = np.ma.empty((radar.nrays, radar.ngates), dtype=float)
    v_vel[:] = np.ma.masked
    w_vel = np.ma.empty((radar.nrays, radar.ngates), dtype=float)
    w_vel[:] = np.ma.masked

    # first guess VAD
    for ind_sweep in range(radar.nsweeps):
        ind_start = radar.sweep_start_ray_index['data'][ind_sweep]
        ind_end = radar.sweep_end_ray_index['data'][ind_sweep]
        for ind_rng in range(radar.ngates):
            vel_azi = vel_aux[ind_start:ind_end+1, ind_rng]

            # check minimum number of valid points
            if npoints_min > 0:
                if vel_azi.compressed().size < npoints_min:
                    continue

            # get position of valid gates
            is_valid_azi = np.logical_not(np.ma.getmaskarray(vel_azi))

            # check maximum allowed gap between data
            if azi_spacing_max > 0.:
                valid_azi = np.sort(
                    radar.azimuth['data'][ind_start:ind_end+1][is_valid_azi])
                delta_azi_max = np.max(np.append(
                    valid_azi[1:]-valid_azi[:-1],
                    valid_azi[0]-(valid_azi[-1]-360)))
                if delta_azi_max > azi_spacing_max:
                    continue

            # get wind coefficients for this azimuth
            u_coeff_aux = u_coeff[ind_start:ind_end+1, ind_rng][is_valid_azi]
            v_coeff_aux = v_coeff[ind_start:ind_end+1, ind_rng][is_valid_azi]
            w_coeff_aux = w_coeff[ind_start:ind_end+1, ind_rng][is_valid_azi]
            coeff_arr = np.array([u_coeff_aux, v_coeff_aux.T, w_coeff_aux]).T

            # retrieve velocity using least square method
            vel_ret, _, _, _ = np.linalg.lstsq(
                coeff_arr, vel_azi.compressed())
            u_vel[ind_start:ind_end+1, ind_rng] = vel_ret[0]
            v_vel[ind_start:ind_end+1, ind_rng] = vel_ret[1]
            w_vel[ind_start:ind_end+1, ind_rng] = vel_ret[2]

    # Compute estimated radial velocity
    vel_est = u_vel*u_coeff+v_vel*v_coeff+w_vel*w_coeff
    if sign == 1:
        vel_est = -vel_est

    return u_vel, v_vel, w_vel, vel_est


def _vel_std(radar, vel, vel_est):
    """
    Computes the variance of the retrieved wind velocity

    Parameters
    ----------
    radar : Radar
        Radar object
    vel : 2D float array
        The measured radial velocity field
    vel_est : 2D float array
        The estimated radial velocity field

    Returns
    -------
    vel_std : 2D float arrays
        The estimated standard deviation at each range gate (one for VAD)
    vel_diff : 2D float array
        The actual velocity difference between estimated and measured radial
        velocities

    """
    vel_diff = vel-vel_est

    vel_std = np.ma.empty((radar.nrays, radar.ngates), dtype=float)
    for ind_sweep in range(radar.nsweeps):
        ind_start = radar.sweep_start_ray_index['data'][ind_sweep]
        ind_end = radar.sweep_end_ray_index['data'][ind_sweep]
        for ind_rng in range(radar.ngates):
            vel_diff_aux = vel_diff[ind_start:ind_end+1, ind_rng]
            nvalid = vel_diff_aux.compressed().size
            vel_std[ind_start:ind_end+1, ind_rng] = np.ma.sqrt(
                np.ma.sum(np.ma.power(vel_diff_aux, 2.))/(nvalid-3))

    return vel_std, vel_diff



""" Unit Tests for Py-ART's graph/gridmapdisplay.py module. """
# execute this script to create figure_gridmapdisplay_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Currently this test only determines if files
# can be created, not that they are correct.

import matplotlib.pyplot as plt
import pytest

import cartopy.crs as ccrs
import pyart

@pytest.mark.skipif(not pyart.graph.gridmapdisplay._XARRAY_AVAILABLE,
                    reason='Xarray is not installed')
def test_gridmapdisplay_simple(outfile=None):
    # test basic GridMapDisplat functionally.
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_grid('reflectivity', vmin=-5, vmax=35, ax=ax)
    if outfile:
        fig.savefig(outfile)

@pytest.mark.skipif(not pyart.graph.gridmapdisplay._XARRAY_AVAILABLE,
                    reason='Xarray is not installed')
def test_gridmapdisplay_imshow(outfile=None):
    # test basic GridMapDisplat functionally.
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_grid('reflectivity', imshow=True,
                      vmin=-5, vmax=35, ax=ax)
    if outfile:
        fig.savefig(outfile)

@pytest.mark.skipif(not pyart.graph.gridmapdisplay._XARRAY_AVAILABLE,
                    reason='Xarray is not installed')
def test_gridmapdisplay_fancy(outfile=None):
    # test a bunch of GridMapDisplay functionaliy
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure(1, figsize=[25., 17.])
    projection = ccrs.Mercator()

    ax1 = plt.subplot(331, projection=projection)
    display.plot_grid('reflectivity', vmin=-5., vmax=35., mask_outside=True,
                      axislabels=('foo', 'bar'), axislabels_flag=True,
                      title='Special title')
    display.plot_crosshairs(color='b')

    ax2 = plt.subplot(332, projection=projection)
    display.plot_grid('reflectivity', axislabels_flag=True)

    ax3 = plt.subplot(333)
    display.plot_colorbar()

    ax4 = plt.subplot(334)
    display.plot_latitude_slice('reflectivity', mask_outside=True)

    ax5 = plt.subplot(335)
    display.plot_latitude_slice('reflectivity', title='Lat title')

    ax6 = plt.subplot(336)
    grid.fields['reflectivity']['valid_min'] = 0
    grid.fields['reflectivity']['valid_max'] = 30
    display.plot_longitude_slice('reflectivity', mask_outside=True)

    ax7 = plt.subplot(337)
    display.plot_longitude_slice('reflectivity', title='Lon title')

    ax8 = plt.subplot(338)
    del display.grid.fields['reflectivity']['long_name']
    display.plot_colorbar()

    if outfile:
        fig.savefig(outfile)

@pytest.mark.skipif(not pyart.graph.gridmapdisplay._XARRAY_AVAILABLE,
                    reason='Xarray is not installed')
def test_generate_filename():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    filename = display.generate_filename('reflectivity', 0)
    assert isinstance(filename, str)

@pytest.mark.skipif(not pyart.graph.gridmapdisplay._XARRAY_AVAILABLE,
                    reason='Xarray is not installed')
def test_generate_titles():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)

    title = display.generate_longitudinal_level_title('reflectivity', 0)
    assert isinstance(title, str)

    title = display.generate_latitudinal_level_title('reflectivity', 0)
    assert isinstance(title, str)

@pytest.mark.skipif(not pyart.graph.gridmapdisplay._XARRAY_AVAILABLE,
                    reason='Xarray is not installed')
def test_error_raising():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)

    # no mappable
    pytest.raises(ValueError, display.plot_colorbar)

    # no field
    display.mappables.append(None) # mock the mappable
    pytest.raises(ValueError, display.plot_colorbar)

if __name__ == '__main__':
    test_gridmapdisplay_simple('figure_grid_mapdisplay_simple.png')
    test_gridmapdisplay_fancy('figure_gridmapdisplay_fancy.png')



"""
pyart.aux_io.rainbow
====================

Routines for reading RAINBOW files (Used by SELEX) using the wradlib library

.. autosummary::
    :toctree: generated/

    read_rainbow_wrl
    _get_angle
    _get_data
    _get_time

"""

# specific modules for this function
import os
from warnings import warn
import datetime

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency

try:
    # `read_rainbow` as of wradlib version 1.0.0
    from wradlib.io import read_Rainbow as read_rainbow
    _WRADLIB_AVAILABLE = True
except ImportError:
    try:
        from wradlib.io import read_rainbow
        _WRADLIB_AVAILABLE = True
    except ImportError:
        _WRADLIB_AVAILABLE = False

RAINBOW_FIELD_NAMES = {
    'W': 'spectrum_width',
    'Wv': 'spectrum_width_vv',  # non standard name
    'Wu': 'unfiltered_spectrum_width',  # non standard name
    'Wvu': 'unfiltered_spectrum_width_vv',  # non standard name
    'V': 'velocity',
    'Vv': 'velocity_vv',  # non standard name
    'Vu': 'unfiltered_velocity',  # non standard name
    'Vvu': 'unfiltered_velocity_vv',  # non standard name
    'dBZ': 'reflectivity',
    'dBZv': 'reflectivity_vv',       # non standard name
    'dBuZ': 'unfiltered_reflectivity',  # non standard name
    'dBuZv': 'unfiltered_reflectivity_vv',  # non standard name
    'ZDR': 'differential_reflectivity',
    'ZDRu': 'unfiltered_differential_reflectivity',  # non standard name
    'RhoHV': 'cross_correlation_ratio',
    'RhoHVu': 'unfiltered_cross_correlation_ratio',  # non standard name
    'PhiDP': 'differential_phase',
    'uPhiDP': 'uncorrected_differential_phase',  # non standard name
    'uPhiDPu':
        'uncorrected_unfiltered_differential_phase',  # non standard name
    'KDP': 'specific_differential_phase',
    'uKDP': 'uncorrected_specific_differential_phase',  # non standard name
    'uKDPu':                                            # non standard name
        'uncorrected_unfiltered_specific_differential_phase',
    'SNRu': 'unfiltered_signal_to_noise_ratio_hh',
    'SNRvu': 'unfiltered_signal_to_noise_ratio_vv',
    'SNR': 'signal_to_noise_ratio_hh',
    'SNRv': 'signal_to_noise_ratio_vv',
    'SQI': 'signal_quality_index',  # non standard name
    'SQIv': 'signal_quality_index_vv',  # non standard name
    'SQIu': 'unfiltered_signal_quality_index',  # non standard name
    'SQIvu': 'unfiltered_signal_quality_index_vv',  # non standard name
    'TEMP': 'temperature',  # non standard name
    'ISO0': 'iso0',  # non standard name
    'VIS': 'visibility'  # non standard name
}

PULSE_WIDTH_VEC = [0.33e-6, 0.5e-6, 1.2e-6, 2.0e-6]  # pulse width [s]


def read_rainbow_wrl(filename, field_names=None, additional_metadata=None,
                     file_field_names=False, exclude_fields=None,
                     include_fields=None, nbytes=4, **kwargs):
    """
    Read a RAINBOW file.
    This routine has been tested to read rainbow5 files version 5.22.3,
    5.34.16 and 5.35.1.
    Since the rainbow file format is evolving constantly there is no guaranty
    that it can work with other versions.
    If necessary, the user should adapt to code according to its own
    file version and raise an issue upstream.

    Data types read by this routine:
    Reflectivity: dBZ, dBuZ, dBZv, dBuZv
    Velocity: V, Vu, Vv, Vvu
    Spectrum width: W, Wu, Wv, Wvu
    Differential reflectivity: ZDR, ZDRu
    Co-polar correlation coefficient: RhoHV, RhoHVu
    Co-polar differential phase: PhiDP, uPhiDP, uPhiDPu
    Specific differential phase: KDP, uKDP, uKDPu
    Signal quality parameters: SQI, SQIu, SQIv, SQIvu
    Temperature: TEMP
    Position of the range bin respect to the ISO0: ISO0
    radar visibility according to Digital Elevation Model (DEM): VIS

    Parameters
    ----------
    filename : str
        Name of the RAINBOW file to read.
    field_names : dict, optional
        Dictionary mapping RAINBOW field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduce any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    nbytes : int
        The number of bytes used to store the data in numpy arrays, e.g. if
        nbytes=4 then floats are going to be stored as np.float32


    Returns
    -------
    radar : Radar
        Radar object containing data from RAINBOW file.

    """
    # check that wradlib is available
    if not _WRADLIB_AVAILABLE:
        raise MissingOptionalDependency(
            "wradlib is required to use read_rainbow_wrl but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # check if it is the right file. Open it and read it
    bfile = os.path.basename(filename)
    supported_file = (bfile.endswith('.vol') or bfile.endswith('.azi') or
                      bfile.endswith('.ele') or bfile.endswith('poi'))
    if not supported_file:
        raise ValueError(
            'Only data files with extension .vol, .azi or .ele are supported')

    if nbytes == 4:
        dtype = np.float32
    elif nbytes == 8:
        dtype = np.float64
    else:
        warn('Number of bytes to store the data ('+str(nbytes) +
             ') not supported. 4 bytes will be used')
        dtype = np.float32

    # create metadata retrieval object
    if field_names is None:
        field_names = RAINBOW_FIELD_NAMES
    filemetadata = FileMetadata('RAINBOW', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    try:
        with open(filename, 'rb') as fid:
            rbf = read_rainbow(fid, loaddata=True)
    except OSError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None

    # check the number of slices
    nslices = int(rbf['volume']['scan']['pargroup']['numele'])
    if nslices > 1:
        single_slice = False
        common_slice_info = rbf['volume']['scan']['slice'][0]
    else:
        single_slice = True
        common_slice_info = rbf['volume']['scan']['slice']

    # check the data type
    # all slices should have the same data type
    datatype = common_slice_info['slicedata']['rawdata']['@type']
    field_name = filemetadata.get_field_name(datatype)
    if field_name is None:
        raise ValueError('Field Name Unknown')

    # get definitions from filemetadata class
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    metadata = filemetadata('metadata')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    elevation = filemetadata('elevation')
    _range = filemetadata('range')
    azimuth = filemetadata('azimuth')
    _time = filemetadata('time')
    field_dic = filemetadata(field_name)

    # other metadata
    frequency = filemetadata('frequency')
    rad_cal_h = filemetadata('calibration_constant_hh')
    rad_cal_v = filemetadata('calibration_constant_vv')
    tx_pwr_h = filemetadata('transmit_power_h')
    tx_pwr_v = filemetadata('transmit_power_v')
    beamwidth_h = filemetadata('radar_beam_width_h')
    beamwidth_v = filemetadata('radar_beam_width_v')
    pulse_width = filemetadata('pulse_width')
    rays_are_indexed = filemetadata('rays_are_indexed')
    ray_angle_res = filemetadata('ray_angle_res')

    # get general file information

    # position and radar frequency
    if 'sensorinfo' in rbf['volume'].keys():
        latitude['data'] = np.array(
            [rbf['volume']['sensorinfo']['lat']], dtype=dtype)
        longitude['data'] = np.array(
            [rbf['volume']['sensorinfo']['lon']], dtype=dtype)
        altitude['data'] = np.array(
            [rbf['volume']['sensorinfo']['alt']], dtype=dtype)
        frequency['data'] = np.array(
            [3e8 / float(rbf['volume']['sensorinfo']['wavelen'])],
            dtype=dtype)
        beamwidth_h['data'] = np.array(
            [float(rbf['volume']['sensorinfo']['beamwidth'])])
        beamwidth_v['data'] = beamwidth_h['data']
    elif 'radarinfo' in rbf['volume'].keys():
        latitude['data'] = np.array(
            [rbf['volume']['radarinfo']['@lat']], dtype=dtype)
        longitude['data'] = np.array(
            [rbf['volume']['radarinfo']['@lon']], dtype=dtype)
        altitude['data'] = np.array(
            [rbf['volume']['radarinfo']['@alt']], dtype=dtype)
        frequency['data'] = np.array(
            [3e8 / float(rbf['volume']['radarinfo']['wavelen'])],
            dtype=dtype)
        beamwidth_h['data'] = np.array(
            [float(rbf['volume']['radarinfo']['beamwidth'])], dtype=dtype)
        beamwidth_v['data'] = beamwidth_h['data']

    # antenna speed
    if 'antspeed' in common_slice_info:
        ant_speed = float(common_slice_info['antspeed'])
    else:
        ant_speed = 10.
        print('WARNING: Unable to read antenna speed. Default value of ' +
              str(ant_speed) + ' deg/s will be used')

    # angle step and sampling mode
    angle_step = float(common_slice_info['anglestep'])
    rays_are_indexed['data'] = None
    ray_angle_res['data'] = None
    if 'fixselect' in common_slice_info:
        if common_slice_info['fixselect'] == 'AngleStep':
            if single_slice:
                rays_are_indexed['data'] = np.array(['true'])
                ray_angle_res['data'] = np.array([angle_step], dtype=dtype)
            else:
                rays_are_indexed['data'] = list()
                ray_angle_res['data'] = np.empty(nslices, dtype=dtype)
                for i in range(nslices):
                    rays_are_indexed['data'].append('true')
                    ray_angle_res['data'][i] = angle_step
                rays_are_indexed['data'] = np.array(rays_are_indexed['data'])
        elif common_slice_info['fixselect'] == 'TimeSamp':
            rays_are_indexed['data'] = list()
            if single_slice:
                rays_are_indexed['data'] = np.array(['false'])
            else:
                for i in range(nslices):
                    rays_are_indexed['data'].append('false')
                rays_are_indexed['data'] = np.array(rays_are_indexed['data'])
        else:
            warn('Unknown sampling mode')

    # sweep_number (is the sweep index)
    sweep_number['data'] = np.arange(nslices, dtype='int32')

    # get number of rays and number of range bins per sweep
    rays_per_sweep = np.empty(nslices, dtype='int32')

    if single_slice:
        rays_per_sweep[0] = int(
            common_slice_info['slicedata']['rawdata']['@rays'])
        nbins = int(common_slice_info['slicedata']['rawdata']['@bins'])
        ssri = np.array([0], dtype='int32')
        seri = np.array([rays_per_sweep[0] - 1], dtype='int32')
    else:
        # number of range bins per ray in sweep
        nbins_sweep = np.empty(nslices, dtype='int32')
        for i in range(nslices):
            slice_info = rbf['volume']['scan']['slice'][i]
            # number of rays per sweep
            rays_per_sweep[i] = int(
                slice_info['slicedata']['rawdata']['@rays'])

            # number of range bins per ray in sweep
            nbins_sweep[i] = int(
                slice_info['slicedata']['rawdata']['@bins'])

        # all sweeps have to have the same number of range bins
        if any(nbins_sweep != nbins_sweep[0]):
            raise ValueError('number of range bins changes between sweeps')
        nbins = nbins_sweep[0]
        ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
        seri = np.cumsum(rays_per_sweep).astype('int32') - 1

    # total number of rays and sweep start ray index and end
    total_rays = sum(rays_per_sweep)
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = seri

    # pulse width and calibration constant
    pulse_width['data'] = None
    rad_cal_h['data'] = None
    rad_cal_v['data'] = None
    tx_pwr_h['data'] = None
    tx_pwr_v['data'] = None
    if 'pw_index' in common_slice_info:
        pw_index = int(common_slice_info['pw_index'])

        pulse_width['data'] = PULSE_WIDTH_VEC[pw_index]*np.ones(
            total_rays, dtype=dtype)

        # calibration constant
        if 'rspdphradconst' in common_slice_info:
            cal_vec = common_slice_info['rspdphradconst'].split()
            rad_cal_h['data'] = np.array(
                [float(cal_vec[pw_index])], dtype=dtype)

        if 'rspdpvradconst' in common_slice_info:
            cal_vec = common_slice_info['rspdpvradconst'].split()
            rad_cal_v['data'] = np.array(
                [float(cal_vec[pw_index])], dtype=dtype)

        # magnetron transmit power
        if 'gdrxmaxpowkw' in common_slice_info:
            tx_pwr_dBm = (
                10.*np.log10(float(common_slice_info['gdrxmaxpowkw'])*1e3)+30.)
            tx_pwr_h['data'] = np.array([tx_pwr_dBm], dtype=dtype)
            tx_pwr_v['data'] = np.array([tx_pwr_dBm], dtype=dtype)

    # range
    r_res = float(common_slice_info['rangestep']) * 1000.
    if 'start_range' in common_slice_info.keys():
        start_range = float(common_slice_info['start_range']) * 1000.
    else:
        start_range = 0.
    _range['data'] = np.linspace(
        start_range+r_res / 2., float(nbins - 1.) * r_res+r_res / 2.,
        nbins).astype(dtype)

    # containers for data
    t_fixed_angle = np.empty(nslices, dtype=dtype)
    moving_angle = np.empty(total_rays, dtype=dtype)
    static_angle = np.empty(total_rays, dtype=dtype)
    time_data = np.empty(total_rays, dtype=dtype)
    fdata = np.ma.zeros((total_rays, nbins), dtype=dtype,
                        fill_value=get_fillvalue())

    # read data from file
    if bfile.endswith('.vol') or bfile.endswith('.azi'):
        scan_type = 'ppi'
        sweep_mode['data'] = np.array(nslices * ['azimuth_surveillance'])
    elif bfile.endswith('.ele'):
        scan_type = 'rhi'
        sweep_mode['data'] = np.array(['elevation_surveillance'])
    else:
        scan_type = 'other'
        sweep_mode['data'] = np.array(['pointing'])

    # read data from file:
    for i in range(nslices):
        if single_slice:
            slice_info = common_slice_info
        else:
            slice_info = rbf['volume']['scan']['slice'][i]

        # fixed angle
        if scan_type == 'other':
            t_fixed_angle[i] = float(slice_info['posazi'])
        else:
            t_fixed_angle[i] = float(slice_info['posangle'])

        # fixed angle (repeated for each ray)
        static_angle[ssri[i]: seri[i]+1] = t_fixed_angle[i]

        # moving angle
        moving_angle[ssri[i]: seri[i]+1], angle_start, angle_stop = (
            _get_angle(slice_info['slicedata']['rayinfo'],
                       angle_step=angle_step, scan_type=scan_type,
                       dtype=dtype))

        # time
        if (isinstance(slice_info['slicedata']['rayinfo'], dict) or
                len(slice_info['slicedata']['rayinfo']) == 2):
            time_data[ssri[i]:seri[i]+1], sweep_start = _get_time(
                slice_info['slicedata']['@date'],
                slice_info['slicedata']['@time'], angle_start[0],
                angle_stop[-1], angle_step, rays_per_sweep[i], ant_speed,
                scan_type=scan_type)
        else:
            sweep_start = datetime.datetime.strptime(
                slice_info['slicedata']['@datetimehighaccuracy'],
                '%Y-%m-%dT%H:%M:%S.%f')
            time_data[ssri[i]:seri[i]+1] = np.array(
                slice_info['slicedata']['rayinfo'][2]['data']*1e-3,
                dtype=np.float64)

        if i == 0:
            start_time = sweep_start
        else:
            time_data[ssri[i]:seri[i]+1] += (
                (sweep_start-start_time).total_seconds())

        # data
        fdata[ssri[i]:seri[i]+1, :] = _get_data(
            slice_info['slicedata']['rawdata'],
            rays_per_sweep[i], nbins, dtype=dtype)

    if bfile.endswith('.vol') or bfile.endswith('.azi'):
        azimuth['data'] = moving_angle
        elevation['data'] = static_angle
    else:
        azimuth['data'] = static_angle
        elevation['data'] = moving_angle

    fixed_angle['data'] = t_fixed_angle

    _time['data'] = time_data
    _time['units'] = make_time_unit_str(start_time)

    # fields
    fields = {}
    # create field dictionary
    field_dic['_FillValue'] = get_fillvalue()
    field_dic['data'] = fdata
    fields[field_name] = field_dic

    # metadata
    # metadata['instrument_name'] = radar_id

    # instrument_parameters
    instrument_parameters = dict()
    instrument_parameters.update({'frequency': frequency})
    instrument_parameters.update({'radar_beam_width_h': beamwidth_h})
    instrument_parameters.update({'radar_beam_width_v': beamwidth_v})
    if pulse_width['data'] is not None:
        instrument_parameters.update({'pulse_width': pulse_width})

    # radar calibration parameters
    radar_calibration = None
    if ((rad_cal_h['data'] is not None) or (rad_cal_v['data'] is not None) or
            (tx_pwr_h['data'] is not None) or (tx_pwr_v['data'] is not None)):
        radar_calibration = dict()
        if rad_cal_h['data'] is not None:
            radar_calibration.update({'calibration_constant_hh': rad_cal_h})
        if rad_cal_v['data'] is not None:
            radar_calibration.update({'calibration_constant_vv': rad_cal_v})
        if tx_pwr_h['data'] is not None:
            radar_calibration.update({'transmit_power_h': tx_pwr_h})
        if tx_pwr_v['data'] is not None:
            radar_calibration.update({'transmit_power_v': tx_pwr_v})

    # angle res
    if rays_are_indexed['data'] is None:
        rays_are_indexed = None
        ray_angle_res = None

    if rays_are_indexed is not None:
        if rays_are_indexed['data'][0] == 'false':
            ray_angle_res = None

    return Radar(_time, _range, fields, metadata, scan_type, latitude,
                 longitude, altitude, sweep_number, sweep_mode, fixed_angle,
                 sweep_start_ray_index, sweep_end_ray_index, azimuth,
                 elevation, rays_are_indexed=rays_are_indexed,
                 ray_angle_res=ray_angle_res,
                 instrument_parameters=instrument_parameters,
                 radar_calibration=radar_calibration)


def _get_angle(ray_info, angle_step=None, scan_type='ppi', dtype=np.float32):
    """
    obtains the ray angle start, stop and center

    Parameters
    ----------
    ray_info : dictionary of dictionaries
        contains the ray info
    angle_step : float
        Optional. The angle step. Used in case there is no information of
        angle stop. Otherwise ignored.
    scan_type : str
        Default ppi. scan_type. Either ppi or rhi.
    dtype : numpy data type object
        The data type of the numpy array where the angles are stored

    Returns
    -------
    moving_angle : numpy array
        the central point of the angle [Deg]
    angle_start :
        the starting point of the angle [Deg]
    angle_stop :
        the end point of the angle [Deg]

    """
    bin_to_deg = 360./65536.

    def _extract_angles(data):
        angle = np.array(data * bin_to_deg, dtype=dtype)
        if scan_type == 'rhi':
            ind = (angle > 225.).nonzero()
            angle[ind] -= 360.
        return angle

    try:
        angle_start = _extract_angles(ray_info['data'])
        if angle_step is None:
            raise ValueError('Unknown angle step')
        angle_stop = angle_start + angle_step
    except TypeError:
        angle_start = _extract_angles(ray_info[0]['data'])
        angle_stop = _extract_angles(ray_info[1]['data'])

    moving_angle = np.angle((np.exp(1.j * np.deg2rad(angle_start)) +
                             np.exp(1.j * np.deg2rad(angle_stop))) / 2.,
                            deg=True)
    moving_angle[moving_angle < 0.] += 360.  # [0, 360]

    return moving_angle, angle_start, angle_stop


def _get_data(rawdata, nrays, nbins, dtype=np.float32):
    """
    Obtains the raw data

    Parameters
    ----------
    rawdata : dictionary of dictionaries
        contains the raw data information
    nrays : int
        Number of rays in sweep
    nbins : int
        Number of bins in ray
    dtype : numpy data type object
        The data type of the numpy array where the data is stored

    Returns
    -------
    data : numpy array
        the data

    """
    databin = rawdata['data']
    datamin = float(rawdata['@min'])
    datamax = float(rawdata['@max'])
    datadepth = float(rawdata['@depth'])
    datatype = rawdata['@type']

    data = np.array(
        datamin+(databin-1)*(datamax-datamin)/(2**datadepth-2),
        dtype=dtype)

    # fill invalid data with fill value
    mask = databin == 0
    data[mask.nonzero()] = get_fillvalue()

    # put phidp data in the range [-180, 180]
    if datatype in ('PhiDP', 'uPhiDP', 'uPhiDPu'):
        is_above_180 = data > 180.
        data[is_above_180.nonzero()] -= 360.

    data = np.reshape(data, [nrays, nbins])
    mask = np.reshape(mask, [nrays, nbins])

    masked_data = np.ma.array(data, mask=mask, fill_value=get_fillvalue())

    return masked_data


def _get_time(date_sweep, time_sweep, first_angle_start, last_angle_stop,
              angle_step, nrays, ant_speed, scan_type='ppi'):
    """
    Computes the time at the center of each ray

    Parameters
    ----------
    date_sweep, time_sweep : str
        the date and time of the sweep
    first_angle_start : float
        The starting point of the first angle in the sweep
    last_angle_stop : float
        The end point of the last angle in the sweep
    nrays : int
        Number of rays in sweep
    ant_speed : float
        antenna speed [deg/s]
    scan_type : str
        Default ppi. scan_type. Either ppi or rhi.

    Returns
    -------
    time_data : numpy array
        the time of each ray since sweep start
    sweep_start : datetime object
        sweep start time

    """
    sweep_start = datetime.datetime.strptime(
        date_sweep+' '+time_sweep, '%Y-%m-%d %H:%M:%S')
    if scan_type in ('ppi', 'other'):
        if (last_angle_stop > first_angle_start) and (
                np.round((last_angle_stop - first_angle_start) /
                         nrays, decimals=2) >= angle_step):
            sweep_duration = (last_angle_stop - first_angle_start) / ant_speed
        elif (last_angle_stop < first_angle_start) and (
                np.round((first_angle_start - last_angle_stop) /
                         nrays, decimals=2) >= angle_step):
            sweep_duration = (first_angle_start - last_angle_stop) / ant_speed
        else:
            sweep_duration = (
                last_angle_stop + 360. - first_angle_start) / ant_speed
    else:
        if last_angle_stop > first_angle_start:
            sweep_duration = (last_angle_stop - first_angle_start) / ant_speed
        else:
            sweep_duration = (first_angle_start - last_angle_stop) / ant_speed

    time_angle = sweep_duration/nrays

    time_data = np.linspace(
        time_angle / 2., sweep_duration-time_angle / 2., num=nrays)

    return time_data, sweep_start



"""
===========================================
Create a 3 panel plot using GridMapDisplay
===========================================

An example that creates a 3 panel plot of a PPI, latitude slice,
and longitude slice using xarray and a cartopy background.

"""

print(__doc__)

# Author: Jason Hemedinger
# License: BSD 3 clause

import cartopy.crs as ccrs
import numpy as np
import matplotlib.pyplot as plt

import pyart

# Read in the gridded file, create GridMapDisplay object
filename = '20110520100000_nexrad_grid.nc'
radar = pyart.io.read_grid(filename)
display = pyart.graph.GridMapDisplay(radar)

# Setting projection, figure size, and panel sizes.
projection = ccrs.PlateCarree()

fig = plt.figure(figsize=[15, 7])

map_panel_axes = [0.05, 0.05, .4, .80]
x_cut_panel_axes = [0.55, 0.10, .4, .25]
y_cut_panel_axes = [0.55, 0.50, .4, .25]

# Set parameters.
level = 1
vmin = -8
vmax = 64
lat = 36.5
lon = -97.7

# Panel 1: PPI plot of the second tilt.
ax1 = fig.add_axes(map_panel_axes, projection=projection)
display.plot_grid('REF', 1, vmin=vmin, vmax=vmax,
                  projection=projection,
                  cmap='pyart_HomeyerRainbow')
display.plot_crosshairs(lon=lon, lat=lat)

# Panel 2: longitude slice
ax2 = fig.add_axes(x_cut_panel_axes)
display.plot_longitude_slice('REF', lon=lon, lat=lat,
                             vmin=vmin, vmax=vmax,
                             cmap='pyart_HomeyerRainbow')

ax2.set_ylim([0, 15])
ax2.set_xlim([-50, 50])

# Panel 3: latitude slice
ax3 = fig.add_axes(y_cut_panel_axes)
display.plot_latitude_slice('REF', lon=lon, lat=lat,
                            vmin=vmin, vmax=vmax,
                            cmap='pyart_HomeyerRainbow')
ax3.set_ylim([0, 15])
ax3.set_xlim([-50, 50])

plt.show()



"""
pyart.core.wind_profile
=======================

Storage of wind profiles.

.. autosummary::
    :toctree: generated/

    HorizontalWindProfile

"""


import numpy as np


class HorizontalWindProfile(object):
    """
    Horizontal wind profile.

    Parameters
    ----------
    height : array-like, 1D
        Heights in meters above sea level at which horizontal winds were
        sampled.
    speed : array-like, 1D
        Horizontal wind speed in meters per second at each height sampled.
    direction : array-like, 1D
        Horizontal wind direction in degrees at each height sampled.

    Other Parameters
    ----------------
    latitude : array-like, 1D, optional
        Latitude in degrees north at each height sampled.
    longitude : array-like, 1D, optional
        Longitude in degrees east at each height sampled.

    Attributes
    ----------
    height : array, 1D
        Heights in meters above sea level at which horizontal winds were
        sampled.
    speed : array, 1D
        Horizontal wind speed in meters per second at each height.
    direction : array, 1D
        Horizontal wind direction in degrees at each height.
    u_wind : array, 1D
        U component of horizontal winds in meters per second at each height.
    v_wind : array, 1D
        V component of horizontal winds in meters per second at each height.

    """

    def __init__(self, height, speed, direction, latitude=None,
                 longitude=None):
        """ initialize """
        if len(height) != len(speed) or len(height) != len(direction):
            raise ValueError("Wind parameters must have the same length.")
        self.height = np.asanyarray(height)
        self.speed = np.asanyarray(speed)
        self.direction = np.asanyarray(direction)
        self._parse_location_data(latitude, longitude)

    @classmethod
    def from_u_and_v(cls, height, u_wind, v_wind):
        """
        Create a HorizontalWindProfile instance from U and V components.

        Parameters
        ----------
        height : array-like, 1D
            Heights in meters above sea level at which horizontal winds were
            sampled.
        u_wind : array-like, 1D
            U component of horizontal wind speed in meters per second.
        v_wind : array-like, 1D
            V component of horizontal wind speed in meters per second.

        """
        u_wind = np.asanyarray(u_wind)
        v_wind = np.asanyarray(v_wind)
        speed = np.sqrt(u_wind*u_wind + v_wind*v_wind)
        direction = np.rad2deg(np.arctan2(-u_wind, -v_wind))
        direction[direction < 0] += 360
        return cls(height, speed, direction)

    @property
    def u_wind(self):
        """ U component of horizontal wind in meters per second. """
        # U = -sin(direction) * speed
        u_wind = -np.sin(np.deg2rad(self.direction)) * self.speed
        return u_wind

    @property
    def v_wind(self):
        """ V component of horizontal wind in meters per second. """
        # V = -cos(direction) * speed
        v_wind = -np.cos(np.deg2rad(self.direction)) * self.speed
        return v_wind

    def _parse_location_data(self, latitude, longitude):
        """ Parse profile location data. """
        if latitude is not None:
            if len(self.height) != len(latitude):
                raise ValueError('Latitude data must have same length.')
            self.latitude = np.asanyarray(latitude)
        if longitude is not None:
            if len(self.height) != len(longitude):
                raise ValueError('Longitude data must have same length.')
            self.longitude = np.asanyarray(longitude)
